# External compatible license
# ******************************************************************************\
# * Copyright (c) 2003-2004, Martin Blais
# * All rights reserved.
# *
# * Redistribution and use in source and binary forms, with or without
# * modification, are permitted provided that the following conditions are
# * met:
# *
# * * Redistributions of source code must retain the above copyright
# *   notice, this list of conditions and the following disclaimer.
# *
# * * Redistributions in binary form must reproduce the above copyright
# *   notice, this list of conditions and the following disclaimer in the
# *   documentation and/or other materials provided with the distribution.
# *
# * * Neither the name of the Martin Blais, Furius, nor the names of its
# *   contributors may be used to endorse or promote products derived from
# *   this software without specific prior written permission.
# *
# * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************\

"""Automatic completion for optparse module.

This module provide automatic bash completion support for programs that use the
optparse module.  The premise is that the optparse options parser specifies
enough information (and more) for us to be able to generate completion strings
esily.  Another advantage of this over traditional completion schemes where the
completion strings are hard-coded in a separate bash source file, is that the
same code that parses the options is used to generate the completions, so the
completions is always up-to-date with the program itself.

In addition, we allow you specify a list of regular expressions or code that
define what kinds of files should be proposed as completions to this file if
needed.  If you want to implement more complex behaviour, you can instead
specify a function, which will be called with the current directory as an
argument.

You need to activate bash completion using the shell script function that comes
with optcomplete (see http://furius.ca/optcomplete for more details).

:author: Martin Blais (blais@furius.ca)
:author: Stijn De Weirdt (Ghent University)

This is a copy of optcomplete.py (changeset 17:e0a9131a94cc)
from source: https://hg.furius.ca/public/optcomplete

Modification by stdweird:
    - cleanup
"""

# Bash Protocol Description
# -------------------------
# 'COMP_CWORD'
#      An index into `${COMP_WORDS}' of the word containing the current
#      cursor position.  This variable is available only in shell
#      functions invoked by the programmable completion facilities (*note
#      Programmable Completion::).
#
# 'COMP_LINE'
#      The current command line.  This variable is available only in
#      shell functions and external commands invoked by the programmable
#      completion facilities (*note Programmable Completion::).
#
# 'COMP_POINT'
#      The index of the current cursor position relative to the beginning
#      of the current command.  If the current cursor position is at the
#      end of the current command, the value of this variable is equal to
#      `${#COMP_LINE}'.  This variable is available only in shell
#      functions and external commands invoked by the programmable
#      completion facilities (*note Programmable Completion::).
#
# 'COMP_WORDS'
#      An array variable consisting of the individual words in the
#      current command line.  This variable is available only in shell
#      functions invoked by the programmable completion facilities (*note
#      Programmable Completion::).
#
# 'COMPREPLY'
#      An array variable from which Bash reads the possible completions
#      generated by a shell function invoked by the programmable
#      completion facility (*note Programmable Completion::).


import copy
import glob
import logging
import os
import re
import shlex
import sys
import types

from optparse import OptionParser, Option
from pprint import pformat

from easybuild.tools.py2vs3 import string_type
from easybuild.tools.utilities import shell_quote

debugfn = None  # for debugging only

OPTCOMPLETE_ENVIRONMENT = 'OPTPARSE_AUTO_COMPLETE'

BASH = "bash"

DEFAULT_SHELL = BASH

SHELL = DEFAULT_SHELL

OPTION_CLASS = Option
OPTIONPARSER_CLASS = OptionParser


def set_optionparser(option_class, optionparser_class):
    """Set the default Option and OptionParser class"""
    global OPTION_CLASS
    global OPTIONPARSER_CLASS
    OPTION_CLASS = option_class
    OPTIONPARSER_CLASS = optionparser_class


def get_shell():
    """Determine the shell, update class constant SHELL and return the shell
    Idea is to call it just once
    """
    global SHELL
    SHELL = os.path.basename(os.environ.get("SHELL", DEFAULT_SHELL))
    return SHELL


# get the shell
get_shell()


class CompleterMissingCallArgument(Exception):
    """Exception to raise when call arg is missing"""


class Completer(object):
    """Base class to derive all other completer classes from.
    It generates an empty completion list
    """
    CALL_ARGS = None  # list of named args that must be passed
    CALL_ARGS_OPTIONAL = None  # list of named args that can be passed

    def __call__(self, **kwargs):
        """Check mandatory args, then return _call"""
        all_args = []
        if self.CALL_ARGS is not None:
            for arg in self.CALL_ARGS:
                all_args.append(arg)
                if arg not in kwargs:
                    msg = "%s __call__ missing mandatory arg %s" % (self.__class__.__name__, arg)
                    raise CompleterMissingCallArgument(msg)

        if self.CALL_ARGS_OPTIONAL is not None:
            all_args.extend(self.CALL_ARGS_OPTIONAL)

        for arg in kwargs.keys():
            if arg not in all_args:
                # remove it
                kwargs.pop(arg)

        return self._call(**kwargs)

    def _call(self, **kwargs):  # pylint: disable=unused-argument
        """Return empty list"""
        return []


class NoneCompleter(Completer):
    """Generates empty completion list. For compatibility reasons."""
    pass


class ListCompleter(Completer):
    """Completes by filtering using a fixed list of strings."""

    def __init__(self, stringlist):
        self.olist = stringlist

    def _call(self, **kwargs):
        """Return the initialised fixed list of strings"""
        return map(str, self.olist)


class AllCompleter(Completer):
    """Completes by listing all possible files in current directory."""
    CALL_ARGS_OPTIONAL = ['pwd']

    def _call(self, **kwargs):
        return os.listdir(kwargs.get('pwd', '.'))


class FileCompleter(Completer):
    """Completes by listing all possible files in current directory.
       If endings are specified, then limit the files to those."""
    CALL_ARGS_OPTIONAL = ['prefix']

    def __init__(self, endings=None):
        if isinstance(endings, string_type):
            endings = [endings]
        elif endings is None:
            endings = []
        self.endings = tuple(map(str, endings))

    def _call(self, **kwargs):
        # TODO : what does prefix do in bash?
        prefix = kwargs.get('prefix', '')

        if SHELL == BASH:
            res = ['_filedir']
            if self.endings:
                res.append("'@(%s)'" % '|'.join(self.endings))
            return " ".join(res)
        else:
            res = []
            for path in glob.glob(prefix + '*'):
                res.append(path)
                if os.path.isdir(path):
                    # add trailing slashes to directories
                    res[-1] += os.path.sep

            if self.endings:
                res = [path for path in res if os.path.isdir(path) or path.endswith(self.endings)]

            if len(res) == 1 and os.path.isdir(res[0]):
                # return two options so that it completes the / but doesn't add a space
                return [res[0] + 'a', res[0] + 'b']
            else:
                return res


class DirCompleter(Completer):
    """Completes by listing subdirectories only."""
    CALL_ARGS_OPTIONAL = ['prefix']

    def _call(self, **kwargs):
        # TODO : what does prefix do in bash?
        prefix = kwargs.get('prefix', '')

        if SHELL == BASH:
            return "_filedir -d"
        else:
            res = [path + "/" for path in glob.glob(prefix + '*') if os.path.isdir(path)]

            if len(res) == 1:
                # return two options so that it completes the / but doesn't add a space
                return [res[0] + 'a', res[0] + 'b']
            else:
                return res


class KnownHostsCompleter(Completer):
    """Completes a list of known hostnames"""

    def _call(self, **kwargs):
        if SHELL == BASH:
            return "_known_hosts"
        else:
            # TODO needs implementation, no autocompletion for now
            return []


class RegexCompleter(Completer):
    """Completes by filtering all possible files with the given list of regexps."""
    CALL_ARGS_OPTIONAL = ['prefix', 'pwd']

    def __init__(self, regexlist, always_dirs=True):
        self.always_dirs = always_dirs

        if isinstance(regexlist, string_type):
            regexlist = [regexlist]
        self.regexlist = []
        for regex in regexlist:
            if isinstance(regex, string_type):
                regex = re.compile(regex)
            self.regexlist.append(regex)

    def _call(self, **kwargs):
        dn = os.path.dirname(kwargs.get('prefix', ''))
        if dn:
            pwd = dn
        else:
            pwd = kwargs.get('pwd', '.')

        ofiles = []
        for fn in os.listdir(pwd):
            for r in self.regexlist:
                if r.match(fn):
                    if dn:
                        fn = os.path.join(dn, fn)
                    ofiles.append(fn)
                    break

            if self.always_dirs and os.path.isdir(fn):
                ofiles.append(fn + os.path.sep)

        return ofiles


class CompleterOption(OPTION_CLASS):
    """optparse Option class with completer attribute"""

    def __init__(self, *args, **kwargs):
        completer = kwargs.pop('completer', None)
        OPTION_CLASS.__init__(self, *args, **kwargs)
        if completer is not None:
            self.completer = completer


def extract_word(line, point):
    """Return a prefix and suffix of the enclosing word.  The character under
    the cursor is the first character of the suffix."""

    if SHELL == BASH and 'IFS' in os.environ:
        ifs = [r.group(0) for r in re.finditer(r'.', os.environ['IFS'])]
        wsre = re.compile('|'.join(ifs))
    else:
        wsre = re.compile(r'\s')

    if point < 0 or point > len(line):
        return '', ''

    preii = point - 1
    while preii >= 0:
        if wsre.match(line[preii]):
            break
        preii -= 1
    preii += 1

    sufii = point
    while sufii < len(line):
        if wsre.match(line[sufii]):
            break
        sufii += 1

    return line[preii:point], line[point:sufii]


def error_override(self, msg):
    """Hack to keep OptionParser from writing to sys.stderr when
    calling self.exit from self.error"""
    self.exit(2, msg=msg)


def guess_first_nonoption(gparser, subcmds_map):
    """Given a global options parser, try to guess the first non-option without
    generating an exception. This is used for scripts that implement a
    subcommand syntax, so that we can generate the appropriate completions for
    the subcommand."""

    gparser = copy.deepcopy(gparser)

    def print_usage_nousage(self, *args, **kwargs):  # pylint: disable=unused-argument
        pass
    gparser.print_usage = print_usage_nousage

    prev_interspersed = gparser.allow_interspersed_args  # save state to restore
    gparser.disable_interspersed_args()

    # interpret cwords like a shell would interpret it
    cwords = shlex.split(os.environ.get('COMP_WORDS', '').strip('() '))

    # save original error_func so we can put it back after the hack
    error_func = gparser.error
    try:
        try:
            instancemethod = type(OPTIONPARSER_CLASS.error)
            # hack to keep OptionParser from writing to sys.stderr
            gparser.error = instancemethod(error_override, gparser, OPTIONPARSER_CLASS)
            _, args = gparser.parse_args(cwords[1:])
        except SystemExit:
            return None
    finally:
        # undo the hack and restore original OptionParser error function
        gparser.error = instancemethod(error_func, gparser, OPTIONPARSER_CLASS)

    value = None
    if args:
        subcmdname = args[0]
        try:
            value = subcmds_map[subcmdname]
        except KeyError:
            pass

    gparser.allow_interspersed_args = prev_interspersed  # restore state

    return value  # can be None, indicates no command chosen.


def autocomplete(parser, arg_completer=None, opt_completer=None, subcmd_completer=None, subcommands=None):
    """Automatically detect if we are requested completing and if so generate
    completion automatically from given parser.

    'parser' is the options parser to use.

    'arg_completer' is a callable object that gets invoked to produce a list of
    completions for arguments completion (oftentimes files).

    'opt_completer' is the default completer to the options that require a
    value.

    'subcmd_completer' is the default completer for the subcommand
    arguments.

    If 'subcommands' is specified, the script expects it to be a map of
    command-name to an object of any kind.  We are assuming that this object is
    a map from command name to a pair of (options parser, completer) for the
    command. If the value is not such a tuple, the method
    'autocomplete(completer)' is invoked on the resulting object.

    This will attempt to match the first non-option argument into a subcommand
    name and if so will use the local parser in the corresponding map entry's
    value.  This is used to implement completion for subcommand syntax and will
    not be needed in most cases.
    """

    # If we are not requested for complete, simply return silently, let the code
    # caller complete. This is the normal path of execution.
    if OPTCOMPLETE_ENVIRONMENT not in os.environ:
        return
    # After this point we should never return, only sys.exit(1)

    # Set default completers.
    if arg_completer is None:
        arg_completer = NoneCompleter()
    if opt_completer is None:
        opt_completer = FileCompleter()
    if subcmd_completer is None:
        # subcmd_completer = arg_completer
        subcmd_completer = FileCompleter()

    # By default, completion will be arguments completion, unless we find out
    # later we're trying to complete for an option.
    completer = arg_completer

    #
    # Completing...
    #

    # Fetching inputs... not sure if we're going to use these.

    # zsh's bashcompinit does not pass COMP_WORDS, replace with
    # COMP_LINE for now...
    if 'COMP_WORDS' not in os.environ:
        os.environ['COMP_WORDS'] = os.environ['COMP_LINE']

    cwords = shlex.split(os.environ.get('COMP_WORDS', '').strip('() '))
    cline = os.environ.get('COMP_LINE', '')
    cpoint = int(os.environ.get('COMP_POINT', 0))
    cword = int(os.environ.get('COMP_CWORD', 0))

    # Extract word enclosed word.
    prefix, suffix = extract_word(cline, cpoint)

    # If requested, try subcommand syntax to find an options parser for that
    # subcommand.
    if subcommands:
        assert isinstance(subcommands, dict)
        value = guess_first_nonoption(parser, subcommands)
        if value:
            if isinstance(value, (list, tuple)):
                parser = value[0]
                if len(value) > 1 and value[1]:
                    # override completer for command if it is present.
                    completer = value[1]
                else:
                    completer = subcmd_completer
                autocomplete(parser, completer)
            elif hasattr(value, 'autocomplete'):
                # Call completion method on object. This should call
                # autocomplete() recursively with appropriate arguments.
                value.autocomplete(subcmd_completer)
            else:
                # no completions for that command object
                pass
            sys.exit(1)
        else:  # suggest subcommands
            completer = ListCompleter(subcommands.keys())

    # Look at previous word, if it is an option and it requires an argument,
    # check for a local completer.  If there is no completer, what follows
    # directly cannot be another option, so mark to not add those to
    # completions.
    optarg = False
    try:
        # Look for previous word, which will be containing word if the option
        # has an equals sign in it.
        prev = None
        if cword < len(cwords):
            mo = re.search('(--.*?)=(.*)', cwords[cword])
            if mo:
                prev, prefix = mo.groups()
        if not prev:
            prev = cwords[cword - 1]

        if prev and prev.startswith('-'):
            option = parser.get_option(prev)
            if option:
                if option.nargs > 0:
                    optarg = True
                    if hasattr(option, 'completer'):
                        completer = option.completer
                    elif option.choices:
                        completer = ListCompleter(option.choices)
                    elif option.type in ('string',):
                        completer = opt_completer
                    else:
                        completer = NoneCompleter()
                # Warn user at least, it could help him figure out the problem.
                elif hasattr(option, 'completer'):
                    msg = "Error: optparse option with a completer does not take arguments: %s" % (option)
                    raise SystemExit(msg)
    except KeyError:
        pass

    completions = []

    # Options completion.
    if not optarg and (not prefix or prefix.startswith('-')):
        completions += parser._short_opt.keys()
        completions += parser._long_opt.keys()
        # Note: this will get filtered properly below.

    completer_kwargs = {
        'pwd': os.getcwd(),
        'cline': cline,
        'cpoint': cpoint,
        'prefix': prefix,
        'suffix': suffix,
    }
    # File completion.
    if completer and (not prefix or not prefix.startswith('-')):
        # Call appropriate completer depending on type.
        if isinstance(completer, (string_type, list, tuple)):
            completer = FileCompleter(completer)
        elif not isinstance(completer, (types.FunctionType, types.LambdaType, types.ClassType, types.ObjectType)):
            # TODO: what to do here?
            pass

        completions = completer(**completer_kwargs)

    if isinstance(completions, string_type):
        # is a bash command, just run it
        if SHELL in (BASH,):  # TODO: zsh
            print(completions)
        else:
            raise Exception("Commands are unsupported by this shell %s" % SHELL)
    else:
        # Filter using prefix.
        if prefix:
            completions = sorted(filter(lambda x: x.startswith(prefix), completions))
        completions = ' '.join(map(str, completions))

        # Save results
        if SHELL == "bash":
            print('COMPREPLY=(' + completions + ')')
        else:
            print(completions)

    # Print debug output (if needed).  You can keep a shell with 'tail -f' to
    # the log file to monitor what is happening.
    if debugfn:
        txt = "\n".join([
            '---------------------------------------------------------',
            'CWORDS %s' % cwords,
            'CLINE %s' % cline,
            'CPOINT %s' % cpoint,
            'CWORD %s' % cword,
            '',
            'Short options',
            pformat(parser._short_opt),
            '',
            'Long options',
            pformat(parser._long_opt),
            'Prefix %s' % prefix,
            'Suffix %s' % suffix,
            'completer_kwargs%s' % str(completer_kwargs),
            # 'completer_completions %s' % completer_completions,
            'completions %s' % completions,
        ])
        if isinstance(debugfn, logging.Logger):
            debugfn.debug(txt)
        else:
            with open(debugfn, 'a') as fh:
                fh.write(txt)

    # Exit with error code (we do not let the caller continue on purpose, this
    # is a run for completions only.)
    sys.exit(1)


class CmdComplete(object):

    """Simple default base class implementation for a subcommand that supports
    command completion.  This class is assuming that there might be a method
    addopts(self, parser) to declare options for this subcommand, and an
    optional completer data member to contain command-specific completion.  Of
    course, you don't really have to use this, but if you do it is convenient to
    have it here."""

    def autocomplete(self, completer=None):
        parser = OPTIONPARSER_CLASS(self.__doc__.strip())
        if hasattr(self, 'addopts'):
            fnc = getattr(self, 'addopts')
            fnc(parser)

        if hasattr(self, 'completer'):
            completer = getattr(self, 'completer')

        return autocomplete(parser, completer)


def gen_cmdline(cmd_list, partial, shebang=True):
    """Create the commandline to generate simulated tabcompletion output
    :param cmd_list: command to execute as list of strings
    :param partial: the string to autocomplete (typically, partial is an element of the cmd_list)
    :param shebang: script has python shebang (if not, add sys.executable)
    """
    cmdline = ' '.join([shell_quote(cmd) for cmd in cmd_list])

    env = []
    env.append("%s=1" % OPTCOMPLETE_ENVIRONMENT)
    env.append('COMP_LINE="%s"' % cmdline)
    env.append('COMP_WORDS="(%s)"' % cmdline)
    env.append("COMP_POINT=%s" % len(cmdline))
    env.append("COMP_CWORD=%s" % cmd_list.index(partial))

    if not shebang:
        env.append(sys.executable)

    # add script
    env.append('"%s"' % cmd_list[0])

    return " ".join(env)



##
# Copyright 2015-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##

"""
Abstract implementation of a package naming scheme.

:author: Robert Schmidt (Ottawa Hospital Research Institute)
:author: Kenneth Hoste (Ghent University)
"""
from abc import ABCMeta, abstractmethod

from easybuild.base import fancylogger
from easybuild.tools.config import build_option


class PackageNamingScheme(object):
    """Abstract class for package naming schemes"""
    __metaclass__ = ABCMeta

    def __init__(self):
        """initialize logger."""
        self.log = fancylogger.getLogger(self.__class__.__name__, fname=False)

    @abstractmethod
    def name(self, ec):
        """Determine package name"""
        pass

    @abstractmethod
    def version(self, ec):
        """Determine package version."""
        pass

    def release(self, ec=None):
        """Determine package release"""
        return build_option('package_release')



# Hooks for HPC2N site changes.
#
# Author: Ake Sandgren, HPC2N

import os

from distutils.version import LooseVersion
from easybuild.framework.easyconfig.format.format import DEPENDENCY_PARAMETERS
from easybuild.tools.filetools import apply_regex_substitutions
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root
from easybuild.tools.systemtools import get_shared_lib_ext


# Add/remove dependencies and/or patches
# Access to the raw values before templating and such.
def parse_hook(ec, *args, **kwargs):

    # Internal helper function
    def add_extra_dependencies(ec, dep_type, extra_deps):
        """dep_type: must be in DEPENDENCY_PARAMETERS or 'osdependencies'"""
        ec.log.info("[parse hook] Adding %s: %s" % (dep_type, extra_deps))

        if dep_type in DEPENDENCY_PARAMETERS:
            for dep in extra_deps:
                ec[dep_type].append(dep)
        elif dep_type == 'osdependencies':
            if isinstance(extra_deps, tuple):
                ec[dep_type].append(extra_deps)
            else:
                raise EasyBuildError("parse_hook: Type of extra_deps argument (%s), for 'osdependencies' must be "
                                     "tuple, found %s" % (extra_deps, type(extra_deps)))
        else:
            raise EasyBuildError("parse_hook: Incorrect dependency type in add_extra_dependencies: %s" % dep_type)

    extra_deps = []

    if ec.name == 'OpenMPI':
        if LooseVersion(ec.version) >= LooseVersion('2') and LooseVersion(ec.version) < LooseVersion('2.1.2'):
            ec.log.info("[parse hook] Adding pmi and lustre patches")
            if LooseVersion(ec.version) < LooseVersion('2.1.1'):
                ec['patches'].append('OpenMPI-2.0.0_fix_bad-include_of_pmi_h.patch')

            if LooseVersion(ec.version) < LooseVersion('2.0.2'):
                ec['patches'].append('OpenMPI-2.0.1_fix_lustre.patch')
            elif LooseVersion(ec.version) < LooseVersion('2.1'):
                ec['patches'].append('OpenMPI-2.0.2_fix_lustre.patch')
            elif LooseVersion(ec.version) < LooseVersion('2.1.1'):
                ec['patches'].append('OpenMPI-2.1.0_fix_lustre.patch')
            else:
                ec['patches'].append('OpenMPI-2.1.1_fix_lustre.patch')

        if LooseVersion(ec.version) == LooseVersion('4.0.0'):
            ec['patches'].append('OpenMPI-4.0.0_fix_configure_bug.patch')

        if LooseVersion(ec.version) >= LooseVersion('2.1'):
            pmix_version = '1.2.5'
            ucx_version = '1.4.0'
            if LooseVersion(ec.version) >= LooseVersion('3'):
                pmix_version = '2.2.1'
            if LooseVersion(ec.version) >= LooseVersion('4'):
                pmix_version = '3.0.2'  # OpenMPI 4.0.0 is not compatible with PMIx 3.1.x

            extra_deps.append(('PMIx', pmix_version))
            # Use of external PMIx requires external libevent
            # But PMIx already has it as a dependency so we don't need
            # to explicitly set it.

            extra_deps.append(('UCX', ucx_version))

    if ec.name == 'impi':
        pmix_version = '3.1.1'
        extra_deps.append(('PMIx', pmix_version))

    if extra_deps:
        add_extra_dependencies(ec, 'dependencies', extra_deps)


def pre_configure_hook(self, *args, **kwargs):
    if self.name == 'GROMACS':
        # HPC2N always uses -DGMX_USE_NVML=ON on GPU builds
        if get_software_root('CUDA'):
            self.log.info("[pre-configure hook] Adding -DGMX_USE_NVML=ON")
            self.cfg.update('configopts', "-DGMX_USE_NVML=ON ")

    if self.name == 'OpenMPI':
        extra_opts = ""
        # Old versions don't work with PMIx, use slurms PMI1
        if LooseVersion(self.version) < LooseVersion('2.1'):
            extra_opts += "--with-pmi=/lap/slurm "
            if LooseVersion(self.version) >= LooseVersion('2'):
                extra_opts += "--with-munge "

        # Using PMIx dependency in easyconfig, see above
        if LooseVersion(self.version) >= LooseVersion('2.1'):
            if get_software_root('PMIx'):
                extra_opts += "--with-pmix=$EBROOTPMIX "
                # Use of external PMIx requires external libevent
                # We're using the libevent that comes from the PMIx dependency
                if get_software_root('libevent'):
                    extra_opts += "--with-libevent=$EBROOTLIBEVENT "
                else:
                    raise EasyBuildError("Error in pre_configure_hook for OpenMPI: External use of PMIx requires "
                                         "external libevent, which was not found. "
                                         "Check parse_hook for dependency settings.")
            else:
                raise EasyBuildError("Error in pre_configure_hook for OpenMPI: PMIx not defined in dependencies. "
                                     "Check parse_hook for dependency settings.")

            if get_software_root('UCX'):
                extra_opts += "--with-ucx=$EBROOTUCX "

        if LooseVersion(self.version) >= LooseVersion('2'):
            extra_opts += "--with-cma "
            extra_opts += "--with-lustre "

        # We still need to fix the knem package to install its
        # pkg-config .pc file correctly, and we need a more generic
        # install dir.
        # extra_opts += "--with-knem=/opt/knem-1.1.2.90mlnx1 "

        self.log.info("[pre-configure hook] Adding %s" % extra_opts)
        self.cfg.update('configopts', extra_opts)

        if LooseVersion(self.version) >= LooseVersion('2.1'):
            self.log.info("[pre-configure hook] Re-enabling ucx")
            self.cfg['configopts'] = self.cfg['configopts'].replace('--without-ucx', ' ')

        self.log.info("[pre-configure hook] Re-enabling dlopen")
        self.cfg['configopts'] = self.cfg['configopts'].replace('--disable-dlopen', ' ')

    if self.name == 'PMIx':
        self.log.info("[pre-configure hook] Adding --with-munge")
        self.cfg.update('configopts', "--with-munge ")
        if LooseVersion(self.version) >= LooseVersion('2'):
            self.log.info("[pre-configure hook] Adding --with-tests-examples")
            self.cfg.update('configopts', "--with-tests-examples ")
            self.log.info("[pre-configure hook] Adding --disable-per-user-config-files")
            self.cfg.update('configopts', "--disable-per-user-config-files")


def pre_build_hook(self, *args, **kwargs):
    if self.name == 'pyslurm':
        self.log.info("[pre-build hook] Adding --slurm=/lap/slurm")
        self.cfg.update('buildopts', "--slurm=/lap/slurm ")


def post_install_hook(self, *args, **kwargs):
    if self.name == 'impi':
        # Fix mpirun from IntelMPI to explicitly unset I_MPI_PMI_LIBRARY
        # it can only be used with srun.
        self.log.info("[post-install hook] Unset I_MPI_PMI_LIBRARY in mpirun")
        apply_regex_substitutions(os.path.join(self.installdir, "intel64", "bin", "mpirun"), [
            (r'^(#!/bin/sh.*)$', r'\1\nunset I_MPI_PMI_LIBRARY'),
        ])


def pre_module_hook(self, *args, **kwargs):
    if self.name == 'impi':
        # Add I_MPI_PMI_LIBRARY to module for IntelMPI so it works with
        # srun.
        self.log.info("[pre-module hook] Set I_MPI_PMI_LIBRARY in impi module")
        # Must be done this way, updating self.cfg['modextravars']
        # directly doesn't work due to templating.
        with self.cfg.disable_templating():
            shlib_ext = get_shared_lib_ext()
            pmix_root = get_software_root('PMIx')
            if pmix_root:
                mpi_type = 'pmix_v3'
                self.cfg['modextravars'].update({
                    'I_MPI_PMI_LIBRARY': os.path.join(pmix_root, "lib", "libpmi." + shlib_ext)
                })
                self.cfg['modextravars'].update({'SLURM_MPI_TYPE': mpi_type})
                # Unfortunately UCX doesn't yet work for unknown reasons. Make sure it is off.
                self.cfg['modextravars'].update({'SLURM_PMIX_DIRECT_CONN_UCX': 'false'})
            else:
                self.cfg['modextravars'].update({'I_MPI_PMI_LIBRARY': "/lap/slurm/lib/libpmi.so"})

    if self.name == 'OpenBLAS':
        self.log.info("[pre-module hook] Set OMP_NUM_THREADS=1 in OpenBLAS module")
        self.cfg.update('modluafooter', 'if ((mode() == "load" and os.getenv("OMP_NUM_THREADS") == nil) '
                        'or (mode() == "unload" and os.getenv("__OpenBLAS_set_OMP_NUM_THREADS") == "1")) then '
                        'setenv("OMP_NUM_THREADS","1"); setenv("__OpenBLAS_set_OMP_NUM_THREADS", "1") end')

    if self.name == 'OpenMPI':
        if LooseVersion(self.version) < LooseVersion('2.1'):
            mpi_type = 'openmpi'
        elif LooseVersion(self.version) < LooseVersion('3'):
            mpi_type = 'pmix_v1'
        elif LooseVersion(self.version) < LooseVersion('4'):
            mpi_type = 'pmix_v2'
        else:
            mpi_type = 'pmix_v3'

        self.log.info("[pre-module hook] Set SLURM_MPI_TYPE=%s in OpenMPI module" % mpi_type)
        # Must be done this way, updating self.cfg['modextravars']
        # directly doesn't work due to templating.
        with self.cfg.disable_templating():
            self.cfg['modextravars'].update({'SLURM_MPI_TYPE': mpi_type})
            # Unfortunately UCX doesn't yet work for unknown reasons. Make sure it is off.
            self.cfg['modextravars'].update({'SLURM_PMIX_DIRECT_CONN_UCX': 'false'})

    if self.name == 'PMIx':
        # This is a, hopefully, temporary workaround for https://github.com/pmix/pmix/issues/1114
        if LooseVersion(self.version) > LooseVersion('2') and LooseVersion(self.version) < LooseVersion('3'):
            self.log.info("[pre-module hook] Set PMIX_MCA_gds=^ds21 in PMIx module")
            with self.cfg.disable_templating():
                self.cfg['modextravars'].update({'PMIX_MCA_gds': '^ds21'})



#!/usr/bin/env python
##
# Copyright 2013-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##

"""
Bootstrap script for EasyBuild

Installs distribute with included (patched) distribute_setup.py script to obtain easy_install,
and then performs a staged install of EasyBuild:
 * stage 0: install setuptools (which provides easy_install), unless already available
 * stage 1: install EasyBuild with easy_install to a temporary directory
 * stage 2: install EasyBuild with EasyBuild from stage 1 to specified install directory

Authors: Kenneth Hoste (UGent), Stijn Deweirdt (UGent), Ward Poelmans (UGent)
License: GPLv2

inspired by https://bitbucket.org/pdubroy/pip/raw/tip/getpip.py
(via http://dubroy.com/blog/so-you-want-to-install-a-python-package/)
"""

import codecs
import copy
import glob
import os
import re
import shutil
import site
import sys
import tempfile
import traceback
from distutils.version import LooseVersion
from hashlib import md5
from platform import python_version

IS_PY3 = sys.version_info[0] == 3

if not IS_PY3:
    import urllib2 as std_urllib
else:
    import urllib.request as std_urllib


EB_BOOTSTRAP_VERSION = '20210715.01'

# argparse preferrred, optparse deprecated >=2.7
HAVE_ARGPARSE = False
try:
    import argparse
    HAVE_ARGPARSE = True
except ImportError:
    import optparse

PYPI_SOURCE_URL = 'https://pypi.python.org/packages/source'

VSC_BASE = 'vsc-base'
VSC_INSTALL = 'vsc-install'
# Python 3 is not supported by the vsc-* packages
EASYBUILD_PACKAGES = (([] if IS_PY3 else [VSC_INSTALL, VSC_BASE]) +
                      ['easybuild-framework', 'easybuild-easyblocks', 'easybuild-easyconfigs'])

STAGE1_SUBDIR = 'eb_stage1'

# the EasyBuild bootstrap script is deprecated, and will only run if $EASYBUILD_BOOTSTRAP_DEPRECATED is defined
EASYBUILD_BOOTSTRAP_DEPRECATED = os.environ.pop('EASYBUILD_BOOTSTRAP_DEPRECATED', None)

# set print_debug to True for detailed progress info
print_debug = os.environ.pop('EASYBUILD_BOOTSTRAP_DEBUG', False)

# install with --force in stage2?
forced_install = os.environ.pop('EASYBUILD_BOOTSTRAP_FORCED', False)

# don't add user site directory to sys.path (equivalent to python -s), see https://www.python.org/dev/peps/pep-0370/
os.environ['PYTHONNOUSERSITE'] = '1'
site.ENABLE_USER_SITE = False

# clean PYTHONPATH to avoid finding readily installed stuff
os.environ['PYTHONPATH'] = ''

EASYBUILD_BOOTSTRAP_SOURCEPATH = os.environ.pop('EASYBUILD_BOOTSTRAP_SOURCEPATH', None)
EASYBUILD_BOOTSTRAP_SKIP_STAGE0 = os.environ.pop('EASYBUILD_BOOTSTRAP_SKIP_STAGE0', False)
EASYBUILD_BOOTSTRAP_FORCE_VERSION = os.environ.pop('EASYBUILD_BOOTSTRAP_FORCE_VERSION', None)

# keep track of original environment (after clearing PYTHONPATH)
orig_os_environ = copy.deepcopy(os.environ)

# If the modules tool is specified, use it
easybuild_modules_tool = os.environ.get('EASYBUILD_MODULES_TOOL', None)
easybuild_module_syntax = os.environ.get('EASYBUILD_MODULE_SYNTAX', None)

# If modules subdir specifications are defined, use them
easybuild_installpath_modules = os.environ.get('EASYBUILD_INSTALLPATH_MODULES', None)
easybuild_subdir_modules = os.environ.get('EASYBUILD_SUBDIR_MODULES', 'modules')
easybuild_suffix_modules_path = os.environ.get('EASYBUILD_SUFFIX_MODULES_PATH', 'all')


#
# Utility functions
#
def debug(msg):
    """Print debug message."""

    if print_debug:
        print("[[DEBUG]] " + msg)


def info(msg):
    """Print info message."""

    print("[[INFO]] " + msg)


def error(msg, exit=True):
    """Print error message and exit."""

    print("[[ERROR]] " + msg)
    sys.exit(1)


def mock_stdout_stderr():
    """Mock stdout/stderr channels"""
    try:
        from cStringIO import StringIO
    except ImportError:
        from io import StringIO
    orig_stdout, orig_stderr = sys.stdout, sys.stderr
    sys.stdout.flush()
    sys.stdout = StringIO()
    sys.stderr.flush()
    sys.stderr = StringIO()

    return orig_stdout, orig_stderr


def restore_stdout_stderr(orig_stdout, orig_stderr):
    """Restore stdout/stderr channels after mocking"""
    # collect output
    sys.stdout.flush()
    stdout = sys.stdout.getvalue()
    sys.stderr.flush()
    stderr = sys.stderr.getvalue()

    # restore original stdout/stderr
    sys.stdout = orig_stdout
    sys.stderr = orig_stderr

    return stdout, stderr


def det_lib_path(libdir):
    """Determine relative path of Python library dir."""
    if libdir is None:
        libdir = 'lib'
    pyver = '.'.join([str(x) for x in sys.version_info[:2]])
    return os.path.join(libdir, 'python%s' % pyver, 'site-packages')


def det_modules_path(install_path):
    """Determine modules path."""
    if easybuild_installpath_modules is not None:
        modules_path = os.path.join(easybuild_installpath_modules, easybuild_suffix_modules_path)
    else:
        modules_path = os.path.join(install_path, easybuild_subdir_modules, easybuild_suffix_modules_path)

    return modules_path


def find_egg_dir_for(path, pkg):
    """Find full path of egg dir for given package."""

    res = None

    for libdir in ['lib', 'lib64']:
        full_libpath = os.path.join(path, det_lib_path(libdir))
        eggdir_regex = re.compile('%s-[0-9a-z.]+-py[0-9.]+.egg' % pkg.replace('-', '_'))
        subdirs = (os.path.exists(full_libpath) and sorted(os.listdir(full_libpath))) or []
        for subdir in subdirs:
            if eggdir_regex.match(subdir):
                eggdir = os.path.join(full_libpath, subdir)
                if res is None:
                    debug("Found egg dir for %s at %s" % (pkg, eggdir))
                    res = eggdir
                else:
                    debug("Found another egg dir for %s at %s (ignoring it)" % (pkg, eggdir))

    # no egg dir found
    if res is None:
        debug("Failed to determine egg dir path for %s in %s (subdirs: %s)" % (pkg, path, subdirs))

    return res


def prep(path):
    """Prepare for installing a Python package in the specified path."""

    debug("Preparing for path %s" % path)

    # restore original environment first
    os.environ = copy.deepcopy(orig_os_environ)
    debug("os.environ['PYTHONPATH'] after reset: %s" % os.environ['PYTHONPATH'])

    # update PATH
    os.environ['PATH'] = os.pathsep.join([os.path.join(path, 'bin')] +
                                         [x for x in os.environ.get('PATH', '').split(os.pathsep) if len(x) > 0])
    debug("$PATH: %s" % os.environ['PATH'])

    # update actual Python search path
    sys.path.insert(0, path)

    # make sure directory exists (this is required by setuptools)
    # usually it's 'lib', but can be 'lib64' as well
    for libdir in ['lib', 'lib64']:
        full_libpath = os.path.join(path, det_lib_path(libdir))
        if not os.path.exists(full_libpath):
            os.makedirs(full_libpath)
        # PYTHONPATH needs to be set as well, otherwise setuptools will fail
        pythonpaths = [x for x in os.environ.get('PYTHONPATH', '').split(os.pathsep) if len(x) > 0]
        os.environ['PYTHONPATH'] = os.pathsep.join([full_libpath] + pythonpaths)

    debug("$PYTHONPATH: %s" % os.environ['PYTHONPATH'])

    os.environ['EASYBUILD_MODULES_TOOL'] = easybuild_modules_tool
    debug("$EASYBUILD_MODULES_TOOL set to %s" % os.environ['EASYBUILD_MODULES_TOOL'])

    if easybuild_module_syntax:
        # if module syntax is specified, use it
        os.environ['EASYBUILD_MODULE_SYNTAX'] = easybuild_module_syntax
        debug("Using specified module syntax: %s" % os.environ['EASYBUILD_MODULE_SYNTAX'])
    elif easybuild_modules_tool != 'Lmod':
        # Lua is the default module syntax, but that requires Lmod
        # if Lmod is not being used, use Tcl module syntax
        os.environ['EASYBUILD_MODULE_SYNTAX'] = 'Tcl'
        debug("$EASYBUILD_MODULE_SYNTAX set to %s" % os.environ['EASYBUILD_MODULE_SYNTAX'])


def check_module_command(tmpdir):
    """Check which module command is available, and prepare for using it."""
    global easybuild_modules_tool

    if easybuild_modules_tool is not None:
        info("Using modules tool specified by $EASYBUILD_MODULES_TOOL: %s" % easybuild_modules_tool)
        return easybuild_modules_tool

    def check_cmd_help(modcmd):
        """Check 'help' output for specified command."""
        modcmd_re = re.compile(r'module\s.*command')
        cmd = "%s python help" % modcmd
        os.system("%s > %s 2>&1" % (cmd, out))
        txt = open(out, 'r').read()
        debug("Output from %s: %s" % (cmd, txt))
        return modcmd_re.search(txt)

    def is_modulecmd_tcl_modulestcl():
        """Determine if modulecmd.tcl is EnvironmentModulesTcl."""
        modcmd_re = re.compile('Modules Release Tcl')
        cmd = "modulecmd.tcl python --version"
        os.system("%s > %s 2>&1" % (cmd, out))
        txt = open(out, 'r').read()
        debug("Output from %s: %s" % (cmd, txt))
        return modcmd_re.search(txt)

    # order matters, which is why we don't use a dict
    known_module_commands = [
        ('lmod', 'Lmod'),
        ('modulecmd.tcl', 'EnvironmentModules'),
        ('modulecmd', 'EnvironmentModulesC'),
    ]
    out = os.path.join(tmpdir, 'module_command.out')
    modtool = None
    for modcmd, modtool in known_module_commands:
        if check_cmd_help(modcmd):
            # distinguish between EnvironmentModulesTcl and EnvironmentModules
            if modcmd == 'modulecmd.tcl' and is_modulecmd_tcl_modulestcl():
                modtool = 'EnvironmentModulesTcl'
            easybuild_modules_tool = modtool
            info("Found module command '%s' (%s), so using it." % (modcmd, modtool))
            break
        elif modcmd == 'lmod':
            # check value of $LMOD_CMD as fallback
            modcmd = os.environ.get('LMOD_CMD')
            if modcmd and check_cmd_help(modcmd):
                easybuild_modules_tool = modtool
                info("Found module command '%s' via $LMOD_CMD (%s), so using it." % (modcmd, modtool))
                break
        elif modtool == 'EnvironmentModules':
            # check value of $MODULESHOME as fallback
            moduleshome = os.environ.get('MODULESHOME', 'MODULESHOME_NOT_DEFINED')
            modcmd = os.path.join(moduleshome, 'libexec', 'modulecmd.tcl')
            if os.path.exists(modcmd) and check_cmd_help(modcmd):
                easybuild_modules_tool = modtool
                info("Found module command '%s' via $MODULESHOME (%s), so using it." % (modcmd, modtool))
                break

    if easybuild_modules_tool is None:
        mod_cmds = [m for (m, _) in known_module_commands]
        msg = [
            "Could not find any module command, make sure one available in your $PATH.",
            "Known module commands are checked in order, and include: %s" % ', '.join(mod_cmds),
            "Check the output of 'type module' to determine the location of the module command you are using.",
        ]
        error('\n'.join(msg))

    return modtool


def check_setuptools():
    """Check whether a suitable setuptools installation is already available."""

    debug("Checking whether suitable setuptools installation is available...")
    res = None

    _, outfile = tempfile.mkstemp()

    # note: we need to be very careful here, because switching to a different setuptools installation (e.g. in stage0)
    #       after the setuptools module was imported is very tricky...
    #       So, we'll check things by running commands through os.system rather than importing setuptools directly.
    cmd_tmpl = "%s -c '%%s' > %s 2>&1" % (sys.executable, outfile)

    # check setuptools version
    try:
        os.system(cmd_tmpl % "import setuptools; print(setuptools.__version__)")
        setuptools_ver = LooseVersion(open(outfile).read().strip())
        debug("Found setuptools version %s" % setuptools_ver)

        min_setuptools_ver = '0.6c11'
        if setuptools_ver < LooseVersion(min_setuptools_ver):
            debug("Minimal setuptools version %s not satisfied, found '%s'" % (min_setuptools_ver, setuptools_ver))
            res = False
    except Exception as err:
        debug("Failed to check setuptools version: %s" % err)
        res = False

    os.system(cmd_tmpl % "from setuptools.command import easy_install; print(easy_install.__file__)")
    out = open(outfile).read().strip()
    debug("Location of setuptools' easy_install module: %s" % out)
    if 'setuptools/command/easy_install' not in out:
        debug("Module 'setuptools.command.easy_install not found")
        res = False

    if res is None:
        os.system(cmd_tmpl % "import setuptools; print(setuptools.__file__)")
        setuptools_loc = open(outfile).read().strip()
        res = os.path.dirname(os.path.dirname(setuptools_loc))
        debug("Location of setuptools installation: %s" % res)

    try:
        os.remove(outfile)
    except Exception:
        pass

    return res


def run_easy_install(args):
    """Run easy_install with specified list of arguments"""
    import setuptools
    debug("Active setuptools installation: %s" % setuptools.__file__)
    from setuptools.command import easy_install

    orig_stdout, orig_stderr = mock_stdout_stderr()
    try:
        easy_install.main(args)
        easy_install_stdout, easy_install_stderr = restore_stdout_stderr(orig_stdout, orig_stderr)
    except (Exception, SystemExit) as err:
        easy_install_stdout, easy_install_stderr = restore_stdout_stderr(orig_stdout, orig_stderr)
        error("Running 'easy_install %s' failed: %s\n%s" % (' '.join(args), err, traceback.format_exc()))

    debug("stdout for 'easy_install %s':\n%s" % (' '.join(args), easy_install_stdout))
    debug("stderr for 'easy_install %s':\n%s" % (' '.join(args), easy_install_stderr))


def check_easy_install_cmd():
    """Try to make sure available 'easy_install' command matches active 'setuptools' installation."""

    debug("Checking whether available 'easy_install' command matches active 'setuptools' installation...")

    _, outfile = tempfile.mkstemp()

    import setuptools
    debug("Location of active setuptools installation: %s" % setuptools.__file__)

    easy_install_regex = re.compile('^(setuptools|distribute) %s' % setuptools.__version__)
    debug("Pattern for 'easy_install --version': %s" % easy_install_regex.pattern)

    pythonpath = os.getenv('PYTHONPATH', '')
    cmd = "PYTHONPATH='%s' %s -m easy_install --version" % (pythonpath, sys.executable)
    os.system("%s > %s 2>&1" % (cmd, outfile))
    outtxt = open(outfile).read().strip()
    debug("Output of '%s':\n%s" % (cmd, outtxt))
    res = bool(easy_install_regex.match(outtxt))
    debug("Result: %s" % res)
    if res:
        debug("Found right 'easy_install' command")
        return

    error("Failed to find right 'easy_install' command!")


#
# Stage functions
#
def stage0(tmpdir):
    """STAGE 0: Prepare and install distribute via included (patched) distribute_setup.py script."""

    print('\n')
    info("+++ STAGE 0: installing distribute via included (patched) distribute_setup.py...\n")

    txt = DISTRIBUTE_SETUP_PY
    if not print_debug:
        # silence distribute_setup.py by redirecting output to /dev/null
        txt = re.sub(r'([^\n]*)(return subprocess.call\(args)(\) == 0)',
                     r"\1f = open(os.devnull, 'w'); \2, stdout=f, stderr=f\3",
                     txt)
        # silence distribute_setup.py completely by setting high log level threshold
        txt = re.sub(r'([^\n]*)(# extracting the tarball[^\n]*)', r'\1log.set_verbosity(1000)\n\1\2', txt)

    # write distribute_setup.py to file (with correct header)
    distribute_setup = os.path.join(tmpdir, 'distribute_setup.py')
    f = open(distribute_setup, "w")
    f.write(txt)
    f.close()

    # create expected directories, set Python search path
    debug("preparing environment...")
    prep(tmpdir)

    import distribute_setup
    debug("distribute_setup.__file__: %s" % distribute_setup.__file__)

    # install easy_install to temporary directory
    from distribute_setup import main as distribute_setup_main
    orig_sys_argv = sys.argv[:]  # make a copy
    sys.argv.append('--prefix=%s' % tmpdir)
    # We download a custom version of distribute: it uses a newer version of markerlib to avoid a bug (#1099)
    # It's is the source of distribute 0.6.49 with the file _markerlib/markers.py replaced by the 0.6 version of
    # markerlib which can be found at https://pypi.python.org/pypi/markerlib/0.6.0
    sys.argv.append('--download-base=https://easybuilders.github.io/easybuild/files/')
    distribute_setup_main(version="0.6.49-patched1")
    sys.argv = orig_sys_argv

    # sanity check
    if os.path.exists(os.path.join(tmpdir, 'bin', 'easy_install')):
        debug("easy_install sanity check OK")
    else:
        error("Installing distribute which should deliver easy_install failed?")

    # prepend distribute egg dir to sys.path, so we know which setuptools we're using
    distribute_egg_dir = find_egg_dir_for(tmpdir, 'distribute')
    if distribute_egg_dir is None:
        error("Failed to determine egg dir path for distribute_egg_dir in %s" % tmpdir)
    else:
        sys.path.insert(0, distribute_egg_dir)

    # make sure we're getting the setuptools we expect
    import setuptools
    from setuptools.command import easy_install

    for mod, path in [('setuptools', setuptools.__file__), ('easy_install', easy_install.__file__)]:
        if tmpdir not in path:
            error("Found another %s module than expected: %s" % (mod, path))
        else:
            debug("Found %s in expected path, good!" % mod)

    info("Installed setuptools version %s (%s)" % (setuptools.__version__, setuptools.__file__))

    return distribute_egg_dir


def stage1(tmpdir, sourcepath, distribute_egg_dir, forcedversion):
    """STAGE 1: temporary install EasyBuild using distribute's easy_install."""

    print('\n')
    info("+++ STAGE 1: installing EasyBuild in temporary dir with easy_install...\n")

    # determine locations of source tarballs, if sources path is specified
    source_tarballs = {}
    if sourcepath is not None:
        info("Fetching sources from %s..." % sourcepath)
        for pkg in EASYBUILD_PACKAGES:
            pkg_tarball_glob = os.path.join(sourcepath, '%s*.tar.gz' % pkg)
            pkg_tarball_paths = glob.glob(pkg_tarball_glob)
            if len(pkg_tarball_paths) > 1:
                error("Multiple tarballs found for %s: %s" % (pkg, pkg_tarball_paths))
            elif len(pkg_tarball_paths) == 0:
                if pkg not in [VSC_BASE, VSC_INSTALL]:
                    # vsc-base package is not strictly required
                    # it's only a dependency since EasyBuild v2.0;
                    # with EasyBuild v2.0, it will be pulled in from PyPI when installing easybuild-framework;
                    # vsc-install is an optional dependency, only required to run unit tests
                    error("Missing source tarball: %s" % pkg_tarball_glob)
            else:
                info("Found %s for %s package" % (pkg_tarball_paths[0], pkg))
                source_tarballs.update({pkg: pkg_tarball_paths[0]})

    if print_debug:
        debug("$ easy_install --help")
        run_easy_install(['--help'])

    # prepare install dir
    targetdir_stage1 = os.path.join(tmpdir, STAGE1_SUBDIR)
    prep(targetdir_stage1)  # set PATH, Python search path

    # install latest EasyBuild with easy_install from PyPi
    cmd = [
        '--upgrade',  # make sure the latest version is pulled from PyPi
        '--prefix=%s' % targetdir_stage1,
    ]

    post_vsc_base = []
    if source_tarballs:
        # install provided source tarballs (order matters)
        cmd.extend([source_tarballs[pkg] for pkg in EASYBUILD_PACKAGES if pkg in source_tarballs])
        # add vsc-base again at the end, to avoid that the one available on the system is used instead
        if VSC_BASE in source_tarballs:
            cmd.append(source_tarballs[VSC_BASE])
    else:
        # install meta-package easybuild from PyPI
        if forcedversion:
            cmd.append('easybuild==%s' % forcedversion)
        elif IS_PY3:
            cmd.append('easybuild>=4.0')  # Python 3 support added in EasyBuild 4
        else:
            cmd.append('easybuild')

        if not IS_PY3:
            # install vsc-base again at the end, to avoid that the one available on the system is used instead
            post_vsc_base = cmd[:]
            post_vsc_base[-1] = VSC_BASE + '<2.9.0'

    if not print_debug:
        cmd.insert(0, '--quiet')

    # There is no support for Python3 in the older vsc-* packages and EasyBuild 4 includes working versions of vsc-*
    if not IS_PY3:
        # install vsc-install version prior to 0.11.4, where mock was introduced as a dependency
        # workaround for problem reported in https://github.com/easybuilders/easybuild-framework/issues/2712
        # also stick to vsc-base < 2.9.0 to avoid requiring 'future' Python package as dependency
        for pkg in [VSC_INSTALL + '<0.11.4', VSC_BASE + '<2.9.0']:
            precmd = cmd[:-1] + [pkg]
            info("running pre-install command 'easy_install %s'" % (' '.join(precmd)))
            run_easy_install(precmd)

    info("installing EasyBuild with 'easy_install %s'\n" % (' '.join(cmd)))
    syntax_error_note = '\n'.join([
        "Note: a 'SyntaxError' may be reported for the easybuild/tools/py2vs3/py%s.py module." % ('3', '2')[IS_PY3],
        "You can safely ignore this message, it will not affect the functionality of the EasyBuild installation.",
        '',
    ])
    info(syntax_error_note)
    run_easy_install(cmd)

    if post_vsc_base:
        info("running post install command 'easy_install %s'" % (' '.join(post_vsc_base)))
        run_easy_install(post_vsc_base)

        pkg_egg_dir = find_egg_dir_for(targetdir_stage1, VSC_BASE)
        if pkg_egg_dir is None:
            # if vsc-base available on system is the same version as the one being installed,
            # the .egg directory may not get installed...
            # in that case, try to have it *copied* by also including --always-copy;
            # using --always-copy should be used as a last resort, since it can result in all kinds of problems
            info(".egg dir for vsc-base not found, trying again with --always-copy...")
            post_vsc_base.insert(0, '--always-copy')
            info("running post install command 'easy_install %s'" % (' '.join(post_vsc_base)))
            run_easy_install(post_vsc_base)

    # clear the Python search path, we only want the individual eggs dirs to be in the PYTHONPATH (see below)
    # this is needed to avoid easy-install.pth controlling what Python packages are actually used
    if distribute_egg_dir is not None:
        os.environ['PYTHONPATH'] = distribute_egg_dir
    else:
        del os.environ['PYTHONPATH']

    # template string to inject in template easyconfig
    templates = {}

    for pkg in EASYBUILD_PACKAGES:
        templates.update({pkg: ''})

        pkg_egg_dir = find_egg_dir_for(targetdir_stage1, pkg)
        if pkg_egg_dir is None:
            if pkg in [VSC_BASE, VSC_INSTALL]:
                # vsc-base is optional in older EasyBuild versions
                continue

        # prepend EasyBuild egg dirs to Python search path, so we know which EasyBuild we're using
        sys.path.insert(0, pkg_egg_dir)
        pythonpaths = [x for x in os.environ.get('PYTHONPATH', '').split(os.pathsep) if len(x) > 0]
        os.environ['PYTHONPATH'] = os.pathsep.join([pkg_egg_dir] + pythonpaths)
        debug("$PYTHONPATH: %s" % os.environ['PYTHONPATH'])

        if source_tarballs:
            if pkg in source_tarballs:
                templates.update({pkg: "'%s'," % os.path.basename(source_tarballs[pkg])})
        else:
            # determine per-package versions based on egg dirs, to use them in easyconfig template
            version_regex = re.compile('%s-([0-9a-z.-]*)-py[0-9.]*.egg' % pkg.replace('-', '_'))
            pkg_egg_dirname = os.path.basename(pkg_egg_dir)
            res = version_regex.search(pkg_egg_dirname)
            if res is not None:
                pkg_version = res.group(1)
                debug("Found version for easybuild-%s: %s" % (pkg, pkg_version))
                templates.update({pkg: "'%s-%s.tar.gz'," % (pkg, pkg_version)})
            else:
                tup = (pkg, pkg_egg_dirname, version_regex.pattern)
                error("Failed to determine version for easybuild-%s package from %s with %s" % tup)

    # figure out EasyBuild version via eb command line
    # note: EasyBuild uses some magic to determine the EasyBuild version based on the versions of the individual pkgs
    ver_regex = {'ver': '[0-9.]*[a-z0-9]*'}
    pattern = r"This is EasyBuild (?P<version>%(ver)s) \(framework: %(ver)s, easyblocks: %(ver)s\)" % ver_regex
    version_re = re.compile(pattern)
    version_out_file = os.path.join(tmpdir, 'eb_version.out')
    eb_version_cmd = 'from easybuild.tools.version import this_is_easybuild; print(this_is_easybuild())'
    cmd = "%s -c '%s' > %s 2>&1" % (sys.executable, eb_version_cmd, version_out_file)
    debug("Determining EasyBuild version using command '%s'" % cmd)
    os.system(cmd)
    txt = open(version_out_file, "r").read()
    res = version_re.search(txt)
    if res:
        eb_version = res.group(1)
        debug("installing EasyBuild v%s" % eb_version)
    else:
        error("Stage 1 failed, could not determine EasyBuild version (txt: %s)." % txt)

    templates.update({'version': eb_version})

    # clear PYTHONPATH before we go to stage2
    # PYTHONPATH doesn't need to (and shouldn't) include the stage1 egg dirs
    os.environ['PYTHONPATH'] = ''

    # make sure we're getting the expected EasyBuild packages
    import easybuild.framework
    import easybuild.easyblocks
    pkgs_to_check = [easybuild.framework, easybuild.easyblocks]
    # vsc is part of EasyBuild 4
    if LooseVersion(eb_version) < LooseVersion('4'):
        import vsc.utils.fancylogger
        pkgs_to_check.append(vsc.utils.fancylogger)

    for pkg in pkgs_to_check:
        if tmpdir not in pkg.__file__:
            error("Found another %s than expected: %s" % (pkg.__name__, pkg.__file__))
        else:
            debug("Found %s in expected path, good!" % pkg.__name__)

    debug("templates: %s" % templates)
    return templates


def stage2(tmpdir, templates, install_path, distribute_egg_dir, sourcepath):
    """STAGE 2: install EasyBuild to temporary dir with EasyBuild from stage 1."""

    print('\n')
    info("+++ STAGE 2: installing EasyBuild in %s with EasyBuild from stage 1...\n" % install_path)

    preinstallopts = ''

    eb_looseversion = LooseVersion(templates['version'])

    # setuptools is no longer required for EasyBuild v4.0 & newer, so skip the setuptools stuff in that case
    if eb_looseversion < LooseVersion('4.0') and distribute_egg_dir is not None:
        # inject path to distribute installed in stage 0 into $PYTHONPATH via preinstallopts
        # other approaches are not reliable, since EasyBuildMeta easyblock unsets $PYTHONPATH;
        # this is required for the easy_install from stage 0 to work
        preinstallopts += "export PYTHONPATH=%s:$PYTHONPATH && " % distribute_egg_dir

        # ensure that (latest) setuptools is installed as well alongside EasyBuild,
        # since it is a required runtime dependency for recent vsc-base and EasyBuild versions
        # this is necessary since we provide our own distribute installation during the bootstrap (cfr. stage0)
        preinstallopts += "%s -m easy_install -U --prefix %%(installdir)s setuptools && " % sys.executable

    # vsc-install is no longer required for EasyBuild v4.0, so skip pre-installed vsc-install in that case
    if eb_looseversion < LooseVersion('4.0'):
        # vsc-install is a runtime dependency for the EasyBuild unit test suite,
        # and is easily picked up from stage1 rather than being actually installed, so force it
        vsc_install = "'%s<0.11.4'" % VSC_INSTALL
        if sourcepath:
            vsc_install_tarball_paths = glob.glob(os.path.join(sourcepath, 'vsc-install*.tar.gz'))
            if len(vsc_install_tarball_paths) == 1:
                vsc_install = vsc_install_tarball_paths[0]
        preinstallopts += "%s -m easy_install -U --prefix %%(installdir)s %s && " % (sys.executable, vsc_install)

    templates.update({
        'preinstallopts': preinstallopts,
    })

    # determine PyPI URLs for individual packages
    pkg_urls = []
    for pkg in EASYBUILD_PACKAGES:

        # vsc-base and vsc-install are not dependencies anymore for EasyBuild v4.0,
        # so skip them here for recent EasyBuild versions
        if eb_looseversion >= LooseVersion('4.0') and pkg in [VSC_INSTALL, VSC_BASE]:
            continue

        # format of pkg entries in templates: "'<pkg_filename>',"
        pkg_filename = templates[pkg][1:-2]

        # the lines below implement a simplified version of the 'pypi_source_urls' and 'derive_alt_pypi_url' functions,
        # which we can't leverage here, partially because of transitional changes in PyPI (#md5= -> #sha256=)

        # determine download URL via PyPI's 'simple' API
        pkg_simple = None
        try:
            pkg_simple = std_urllib.urlopen('https://pypi.python.org/simple/%s' % pkg, timeout=10).read()
        except (std_urllib.URLError, std_urllib.HTTPError) as err:
            # failing to figure out the package download URl may be OK when source tarballs are provided
            if sourcepath:
                info("Ignoring failed attempt to determine '%s' download URL since source tarballs are provided" % pkg)
            else:
                raise err

        if pkg_simple:
            if IS_PY3:
                pkg_simple = pkg_simple.decode('utf-8')
            pkg_url_part_regex = re.compile('/(packages/[^#]+)/%s#' % pkg_filename)
            res = pkg_url_part_regex.search(pkg_simple)
            if res:
                pkg_url = 'https://pypi.python.org/' + res.group(1)
                pkg_urls.append(pkg_url)
            elif sourcepath:
                info("Ignoring failure to determine source URL for '%s' (source tarballs are provided)" % pkg_filename)
            else:
                error_msg = "Failed to determine PyPI package URL for %s using pattern '%s': %s\n"
                error(error_msg % (pkg, pkg_url_part_regex.pattern, pkg_simple))

    # vsc-base and vsc-install are no longer required for EasyBuild v4.0.0,
    # so only include them in 'sources' for older versions
    sources_tmpl = "%(easybuild-framework)s%(easybuild-easyblocks)s%(easybuild-easyconfigs)s"
    if eb_looseversion < LooseVersion('4.0'):
        sources_tmpl = "%(vsc-install)s%(vsc-base)s" + sources_tmpl

    templates.update({
        'source_urls': '\n'.join(["'%s'," % x for x in pkg_urls]),
        'sources': sources_tmpl % templates,
        'pythonpath': distribute_egg_dir,
    })

    # create easyconfig file
    ebfile = os.path.join(tmpdir, 'EasyBuild-%s.eb' % templates['version'])
    handle = open(ebfile, 'w')
    ebfile_txt = EASYBUILD_EASYCONFIG_TEMPLATE % templates
    handle.write(ebfile_txt)
    handle.close()
    debug("Contents of generated easyconfig file:\n%s" % ebfile_txt)

    # set command line arguments for eb
    eb_args = ['eb', ebfile, '--allow-modules-tool-mismatch']
    if print_debug:
        eb_args.extend(['--debug', '--logtostdout'])
    if forced_install:
        info("Performing FORCED installation, as requested...")
        eb_args.append('--force')

    # make sure we don't leave any stuff behind in default path $HOME/.local/easybuild
    # and set build and install path explicitely
    if LooseVersion(templates['version']) < LooseVersion('1.3.0'):
        os.environ['EASYBUILD_PREFIX'] = tmpdir
        os.environ['EASYBUILD_BUILDPATH'] = tmpdir
        if install_path is not None:
            os.environ['EASYBUILD_INSTALLPATH'] = install_path
    else:
        # only for v1.3 and up
        eb_args.append('--prefix=%s' % tmpdir)
        eb_args.append('--buildpath=%s' % tmpdir)
        if install_path is not None:
            eb_args.append('--installpath=%s' % install_path)
        if sourcepath is not None:
            eb_args.append('--sourcepath=%s' % sourcepath)

        # make sure EasyBuild can find EasyBuild-*.eb easyconfig file when it needs to;
        # (for example when HierarchicalMNS is used as module naming scheme,
        #  see https://github.com/easybuilders/easybuild-framework/issues/2393)
        eb_args.append('--robot-paths=%s:' % tmpdir)

    # make sure parent modules path already exists (Lmod trips over a non-existing entry in $MODULEPATH)
    if install_path is not None:
        modules_path = det_modules_path(install_path)
        if not os.path.exists(modules_path):
            os.makedirs(modules_path)
        debug("Created path %s" % modules_path)

    debug("Running EasyBuild with arguments '%s'" % ' '.join(eb_args))
    sys.argv = eb_args

    # location to 'eb' command (from stage 1) may be expected to be included in $PATH
    # it usually is there after stage1, unless 'prep' is called again with another location
    # (only when stage 0 is not skipped)
    # cfr. https://github.com/easybuilders/easybuild-framework/issues/2279
    curr_path = [x for x in os.environ.get('PATH', '').split(os.pathsep) if len(x) > 0]
    os.environ['PATH'] = os.pathsep.join([os.path.join(tmpdir, STAGE1_SUBDIR, 'bin')] + curr_path)
    debug("$PATH: %s" % os.environ['PATH'])

    # install EasyBuild with EasyBuild
    from easybuild.main import main as easybuild_main
    easybuild_main()

    if print_debug:
        os.environ['EASYBUILD_DEBUG'] = '1'

    # make sure the EasyBuild module was actually installed
    # EasyBuild configuration options that are picked up from configuration files/environment may break the bootstrap,
    # for example by having $EASYBUILD_VERSION defined or via a configuration file specifies a value for 'stop'...
    from easybuild.tools.config import build_option, install_path, get_module_syntax
    from easybuild.framework.easyconfig.easyconfig import ActiveMNS
    eb_spec = {
        'name': 'EasyBuild',
        'hidden': False,
        'toolchain': {'name': 'dummy', 'version': 'dummy'},
        'version': templates['version'],
        'versionprefix': '',
        'versionsuffix': '',
        'moduleclass': 'tools',
    }

    mod_path = os.path.join(install_path('mod'), build_option('suffix_modules_path'))
    debug("EasyBuild module should have been installed to %s" % mod_path)

    eb_mod_name = ActiveMNS().det_full_module_name(eb_spec)
    debug("EasyBuild module name: %s" % eb_mod_name)

    eb_mod_path = os.path.join(mod_path, eb_mod_name)
    if get_module_syntax() == 'Lua':
        eb_mod_path += '.lua'

    if os.path.exists(eb_mod_path):
        info("EasyBuild module installed: %s" % eb_mod_path)
    else:
        error("EasyBuild module not found at %s, define $EASYBUILD_BOOTSTRAP_DEBUG to debug" % eb_mod_path)


def main():
    """Main script: bootstrap EasyBuild in stages."""

    self_txt = open(__file__).read()
    if IS_PY3:
        self_txt = self_txt.encode('utf-8')

    url = 'https://docs.easybuild.io/en/latest/Installation.html'
    info("Use of the EasyBuild boostrap script is DEPRECATED (since June 2021).")
    info("It is strongly recommended to use one of the installation methods outlined at %s instead!\n" % url)
    if not EASYBUILD_BOOTSTRAP_DEPRECATED:
        error("The EasyBuild bootstrap script will only run if $EASYBUILD_BOOTSTRAP_DEPRECATED is defined.")
    else:
        msg = "You have opted to continue with the EasyBuild bootstrap script by defining "
        msg += "$EASYBUILD_BOOTSTRAP_DEPRECATED. Good luck!\n"
        info(msg)

    info("EasyBuild bootstrap script (version %s, MD5: %s)" % (EB_BOOTSTRAP_VERSION, md5(self_txt).hexdigest()))
    info("Found Python %s\n" % '; '.join(sys.version.split('\n')))

    # disallow running as root, since stage 2 will fail
    if os.getuid() == 0:
        error("Don't run the EasyBuild bootstrap script as root, "
              "since stage 2 (installing EasyBuild with EasyBuild) will fail.")

    # general option/argument parser
    if HAVE_ARGPARSE:
        bs_argparser = argparse.ArgumentParser()
        bs_argparser.add_argument("prefix", help="Installation prefix directory",
                                  type=str)
        bs_args = bs_argparser.parse_args()

        # prefix specification
        install_path = os.path.abspath(bs_args.prefix)
    else:
        bs_argparser = optparse.OptionParser(usage="usage: %prog [options] prefix")
        (bs_opts, bs_args) = bs_argparser.parse_args()

        # poor method, but should prefer argparse module for better pos arg support.
        if len(bs_args) < 1:
            error("Too few arguments\n" + bs_argparser.get_usage())
        elif len(bs_args) > 1:
            error("Too many arguments\n" + bs_argparser.get_usage())

        # prefix specification
        install_path = os.path.abspath(str(bs_args[0]))

    info("Installation prefix %s" % install_path)

    sourcepath = EASYBUILD_BOOTSTRAP_SOURCEPATH
    if sourcepath is not None:
        info("Fetching sources from %s..." % sourcepath)

    forcedversion = EASYBUILD_BOOTSTRAP_FORCE_VERSION
    if forcedversion:
        info("Forcing specified version %s..." % forcedversion)
        if IS_PY3 and LooseVersion(forcedversion) < LooseVersion('4'):
            error('Python 3 support is only available with EasyBuild 4.x but you are trying to install EasyBuild %s'
                  % forcedversion)

    # create temporary dir for temporary installations
    tmpdir = tempfile.mkdtemp()
    debug("Going to use %s as temporary directory" % tmpdir)
    os.chdir(tmpdir)

    # check whether a module command is available, we need that
    modtool = check_module_command(tmpdir)

    # clean sys.path, remove paths that may contain EasyBuild packages or stuff installed with easy_install
    orig_sys_path = sys.path[:]
    sys.path = []
    for path in orig_sys_path:
        include_path = True
        # exclude path if it's potentially an EasyBuild/VSC package, providing the 'easybuild'/'vsc' namespace, resp.
        if any(os.path.exists(os.path.join(path, pkg, '__init__.py')) for pkg in ['easyblocks', 'easybuild', 'vsc']):
            include_path = False
        # exclude any .egg paths
        if path.endswith('.egg'):
            include_path = False
        # exclude any path that contains an easy-install.pth file
        if os.path.exists(os.path.join(path, 'easy-install.pth')):
            include_path = False

        if include_path:
            sys.path.append(path)
        else:
            debug("Excluding %s from sys.path" % path)

    debug("sys.path after cleaning: %s" % sys.path)

    # install EasyBuild in stages

    # STAGE 0: install distribute, which delivers easy_install
    distribute_egg_dir = None
    if EASYBUILD_BOOTSTRAP_SKIP_STAGE0:
        info("Skipping stage 0, using local distribute/setuptools providing easy_install")
    else:
        setuptools_loc = check_setuptools()
        if setuptools_loc:
            info("Suitable setuptools installation already found, skipping stage 0...")
            sys.path.insert(0, setuptools_loc)
        else:
            info("No suitable setuptools installation found, proceeding with stage 0...")
            distribute_egg_dir = stage0(tmpdir)

    # STAGE 1: install EasyBuild using easy_install to tmp dir
    templates = stage1(tmpdir, sourcepath, distribute_egg_dir, forcedversion)

    # add location to easy_install provided through stage0 to $PATH
    # this must be done *after* stage1, since $PATH is reset during stage1
    if distribute_egg_dir:
        prep(tmpdir)

    # make sure the active 'easy_install' is the right one (i.e. it matches the active setuptools installation)
    check_easy_install_cmd()

    # STAGE 2: install EasyBuild using EasyBuild (to final target installation dir)
    stage2(tmpdir, templates, install_path, distribute_egg_dir, sourcepath)

    # clean up the mess
    debug("Cleaning up %s..." % tmpdir)
    shutil.rmtree(tmpdir)

    print('')
    info('Bootstrapping EasyBuild completed!\n')

    if install_path is not None:
        info('EasyBuild v%s was installed to %s, so make sure your $MODULEPATH includes %s' %
             (templates['version'], install_path, det_modules_path(install_path)))
    else:
        info('EasyBuild v%s was installed to configured install path, make sure your $MODULEPATH is set correctly.' %
             templates['version'])
        info('(default config => add "$HOME/.local/easybuild/modules/all" in $MODULEPATH)')

    print('')
    info("Run 'module load EasyBuild', and run 'eb --help' to get help on using EasyBuild.")
    info("Set $EASYBUILD_MODULES_TOOL to '%s' to use the same modules tool as was used now." % modtool)
    print('')
    info("By default, EasyBuild will install software to $HOME/.local/easybuild.")
    info("To install software with EasyBuild to %s, set $EASYBUILD_INSTALLPATH accordingly." % install_path)
    info("See http://easybuild.readthedocs.org/en/latest/Configuration.html for details on configuring EasyBuild.")


# template easyconfig file for EasyBuild
EASYBUILD_EASYCONFIG_TEMPLATE = """
easyblock = 'EB_EasyBuildMeta'

name = 'EasyBuild'
version = '%(version)s'

homepage = 'http://easybuilders.github.com/easybuild/'
description = \"\"\"EasyBuild is a software build and installation framework
written in Python that allows you to install software in a structured,
repeatable and robust way.\"\"\"

toolchain = {'name': 'dummy', 'version': 'dummy'}

source_urls = [%(source_urls)s]
sources = [%(sources)s]

# EasyBuild is a (set of) Python packages, so it depends on Python
# usually, we want to use the system Python, so no actual Python dependency is listed
allow_system_deps = [('Python', SYS_PYTHON_VERSION)]

preinstallopts = "%(preinstallopts)s"

sanity_check_paths = {
    'files': ['bin/eb'],
    'dirs': ['lib'],
}

moduleclass = 'tools'
"""

# check Python version
loose_pyver = LooseVersion(python_version())
min_pyver2 = LooseVersion('2.6')
min_pyver3 = LooseVersion('3.5')
if loose_pyver < min_pyver2 or (loose_pyver >= LooseVersion('3') and loose_pyver < min_pyver3):
    sys.stderr.write("ERROR: Incompatible Python version: %s (should be Python 2 >= %s or Python 3 >= %s)\n"
                     % (python_version(), min_pyver2, min_pyver3))
    sys.exit(1)

# distribute_setup.py script (https://pypi.python.org/pypi/distribute)
#
# A compressed copy of a patched distribute_setup.py (version 0.6.49), generated like so:
# >>> import base64
# >>> import zlib
# >>> base64.b64encode(zlib.compress(open("distribute_setup.py").read()))
# compressed copy below is for setuptools 0.6c11, after applying patch:
#
# --- distribute_setup.py.orig	2013-07-05 03:50:13.000000000 +0200
# +++ distribute_setup.py	2015-11-27 12:20:12.040032041 +0100
# @@ -528,6 +528,8 @@
#              log.warn("--user requires Python 2.6 or later")
#              raise SystemExit(1)
#          install_args.append('--user')
# +    if options.prefix_install:
# +        install_args.append('--prefix=%s' % options.prefix_install)
#      return install_args
#
#  def _parse_args():
# @@ -539,6 +541,8 @@
#          '--user', dest='user_install', action='store_true', default=False,
#          help='install in user site package (requires Python 2.6 or later)')
#      parser.add_option(
# +        '--prefix', dest='prefix_install', metavar="PATH", help='install in prefix')
# +    parser.add_option(
#          '--download-base', dest='download_base', metavar="URL",
#          default=DEFAULT_URL,
#          help='alternative URL from where to download the distribute package')
# @@ -549,7 +553,7 @@
#  def main(version=DEFAULT_VERSION):
#      """Install or upgrade setuptools and EasyInstall"""
#      options = _parse_args()
# -    tarball = download_setuptools(download_base=options.download_base)
# +    tarball = download_setuptools(version=version, download_base=options.download_base)
#      return _install(tarball, _build_install_args(options))
#
#  if __name__ == '__main__':
#
DISTRIBUTE_SETUP_PY = """

7lwukTHLpfpLDH2GT+yCCf8D2cp1xw==
"""
if IS_PY3:
    DISTRIBUTE_SETUP_PY = DISTRIBUTE_SETUP_PY.encode('ascii')
DISTRIBUTE_SETUP_PY = codecs.decode(codecs.decode(DISTRIBUTE_SETUP_PY, "base64"), "zlib")

# run main function as body of script
main()



# #
# Copyright 2013-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #
"""
Unit tests for eb command line options.

@author: Kenneth Hoste (Ghent University)
"""
import os
import sys
from test.framework.utilities import EnhancedTestCase, TestLoaderFiltered
from unittest import TextTestRunner

from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import mkdir, write_file
from easybuild.tools.include import include_easyblocks, include_module_naming_schemes, include_toolchains
from easybuild.tools.include import is_software_specific_easyblock


def up(path, cnt):
    """Return path N times up."""
    if cnt > 0:
        path = up(os.path.dirname(path), cnt - 1)
    return path


class IncludeTest(EnhancedTestCase):
    """Testcases for command line options."""

    logfile = None

    def test_include_easyblocks(self):
        """Test include_easyblocks()."""
        test_easyblocks = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sandbox', 'easybuild', 'easyblocks')

        # put a couple of custom easyblocks in place, to test
        myeasyblocks = os.path.join(self.test_prefix, 'myeasyblocks')
        mkdir(os.path.join(myeasyblocks, 'generic'), parents=True)

        # include __init__.py files that should be ignored, and shouldn't cause trouble (bug #1697)
        write_file(os.path.join(myeasyblocks, '__init__.py'), "# dummy init, should not get included")
        write_file(os.path.join(myeasyblocks, 'generic', '__init__.py'), "# dummy init, should not get included")

        myfoo_easyblock_txt = '\n'.join([
            "from easybuild.easyblocks.generic.configuremake import ConfigureMake",
            "class EB_Foo(ConfigureMake):",
            "   pass",
        ])
        write_file(os.path.join(myeasyblocks, 'myfoo.py'), myfoo_easyblock_txt)

        mybar_easyblock_txt = '\n'.join([
            "from easybuild.framework.easyblock import EasyBlock",
            "class Bar(EasyBlock):",
            "   pass",
        ])
        write_file(os.path.join(myeasyblocks, 'generic', 'mybar.py'), mybar_easyblock_txt)

        # second myfoo easyblock, should get ignored...
        myfoo_bis = os.path.join(self.test_prefix, 'myfoo.py')
        write_file(myfoo_bis, '')

        # hijack $HOME to test expanding ~ in locations passed to include_easyblocks
        os.environ['HOME'] = myeasyblocks

        # expand set of known easyblocks with our custom ones;
        # myfoo easyblock is included twice, first path should have preference
        glob_paths = [os.path.join('~', '*'), os.path.join(myeasyblocks, '*/*.py'), myfoo_bis]
        included_easyblocks_path = include_easyblocks(self.test_prefix, glob_paths)

        expected_paths = ['__init__.py', 'easyblocks/__init__.py', 'easyblocks/myfoo.py',
                          'easyblocks/generic/__init__.py', 'easyblocks/generic/mybar.py']
        for filepath in expected_paths:
            fullpath = os.path.join(included_easyblocks_path, 'easybuild', filepath)
            self.assertTrue(os.path.exists(fullpath), "%s exists" % fullpath)

        # path to included easyblocks should be prepended to Python search path
        self.assertEqual(sys.path[0], included_easyblocks_path)

        # importing custom easyblocks should work
        import easybuild.easyblocks.myfoo
        myfoo_pyc_path = easybuild.easyblocks.myfoo.__file__
        myfoo_real_py_path = os.path.realpath(os.path.join(os.path.dirname(myfoo_pyc_path), 'myfoo.py'))
        self.assertTrue(os.path.samefile(up(myfoo_real_py_path, 1), myeasyblocks))
        del sys.modules['easybuild.easyblocks.myfoo']

        import easybuild.easyblocks.generic.mybar
        mybar_pyc_path = easybuild.easyblocks.generic.mybar.__file__
        mybar_real_py_path = os.path.realpath(os.path.join(os.path.dirname(mybar_pyc_path), 'mybar.py'))
        self.assertTrue(os.path.samefile(up(mybar_real_py_path, 2), myeasyblocks))
        del sys.modules['easybuild.easyblocks.generic.mybar']

        # existing (test) easyblocks are unaffected
        import easybuild.easyblocks.foofoo
        foofoo_path = os.path.dirname(os.path.dirname(easybuild.easyblocks.foofoo.__file__))
        self.assertTrue(os.path.samefile(foofoo_path, test_easyblocks))
        del sys.modules['easybuild.easyblocks.foofoo']

    def test_include_easyblocks_priority(self):
        """Test whether easyblocks included via include_easyblocks() get priority over others."""
        test_easyblocks = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sandbox', 'easybuild', 'easyblocks')

        # make sure that test 'foo' easyblock is there
        import easybuild.easyblocks.foo
        foo_path = os.path.dirname(os.path.dirname(easybuild.easyblocks.foo.__file__))
        self.assertTrue(os.path.samefile(foo_path, test_easyblocks))

        # inject custom 'foo' easyblocks
        myeasyblocks = os.path.join(self.test_prefix, 'myeasyblocks')
        mkdir(myeasyblocks)

        # include __init__.py file that should be ignored, and shouldn't cause trouble (bug #1697)
        write_file(os.path.join(myeasyblocks, '__init__.py'), "# dummy init, should not get included")

        # 'undo' import of foo easyblock
        del sys.modules['easybuild.easyblocks.foo']

        foo_easyblock_txt = '\n'.join([
            "from easybuild.framework.easyblock import EasyBlock",
            "class EB_Foo(EasyBlock):",
            "   pass",
        ])
        write_file(os.path.join(myeasyblocks, 'foo.py'), foo_easyblock_txt)

        # check that the sandboxed easyblock is imported before include_easyblocks is run
        foo_pyc_path = easybuild.easyblocks.foo.__file__
        foo_real_py_path = os.path.realpath(os.path.join(os.path.dirname(foo_pyc_path), 'foo.py'))
        self.assertTrue(os.path.samefile(os.path.dirname(os.path.dirname(foo_pyc_path)), test_easyblocks))
        self.assertFalse(os.path.samefile(foo_real_py_path, os.path.join(myeasyblocks, 'foo.py')))

        include_easyblocks(self.test_prefix, [os.path.join(myeasyblocks, 'foo.py')])

        # check that the included easyblock is imported after include_easyblocks is run
        foo_pyc_path = easybuild.easyblocks.foo.__file__
        foo_real_py_path = os.path.realpath(os.path.join(os.path.dirname(foo_pyc_path), 'foo.py'))
        self.assertFalse(os.path.samefile(os.path.dirname(os.path.dirname(foo_pyc_path)), test_easyblocks))
        self.assertTrue(os.path.samefile(foo_real_py_path, os.path.join(myeasyblocks, 'foo.py')))

        # 'undo' import of foo easyblock
        del sys.modules['easybuild.easyblocks.foo']

    def test_include_mns(self):
        """Test include_module_naming_schemes()."""

        my_mns = os.path.join(self.test_prefix, 'my_mns')
        mkdir(my_mns)

        # include __init__.py file that should be ignored, and shouldn't cause trouble (bug #1697)
        write_file(os.path.join(my_mns, '__init__.py'), "# dummy init, should not get included")

        my_mns_txt = '\n'.join([
            "from easybuild.tools.module_naming_scheme.mns import ModuleNamingScheme",
            "class MyMNS(ModuleNamingScheme):",
            "   pass",
        ])
        write_file(os.path.join(my_mns, 'my_mns.py'), my_mns_txt)

        my_mns_bis = os.path.join(self.test_prefix, 'my_mns.py')
        write_file(my_mns_bis, '')

        # include custom MNS
        included_mns_path = include_module_naming_schemes(self.test_prefix, [os.path.join(my_mns, '*.py'), my_mns_bis])

        expected_paths = ['__init__.py', 'tools/__init__.py', 'tools/module_naming_scheme/__init__.py',
                          'tools/module_naming_scheme/my_mns.py']
        for filepath in expected_paths:
            fullpath = os.path.join(included_mns_path, 'easybuild', filepath)
            self.assertTrue(os.path.exists(fullpath), "%s exists" % fullpath)

        # path to included MNSs should be prepended to Python search path
        self.assertEqual(sys.path[0], included_mns_path)

        # importing custom MNS should work
        import easybuild.tools.module_naming_scheme.my_mns
        my_mns_pyc_path = easybuild.tools.module_naming_scheme.my_mns.__file__
        my_mns_real_py_path = os.path.realpath(os.path.join(os.path.dirname(my_mns_pyc_path), 'my_mns.py'))
        self.assertTrue(os.path.samefile(up(my_mns_real_py_path, 1), my_mns))

    def test_include_toolchains(self):
        """Test include_toolchains()."""
        my_toolchains = os.path.join(self.test_prefix, 'my_toolchains')
        mkdir(my_toolchains)

        # include __init__.py file that should be ignored, and shouldn't cause trouble (bug #1697)
        write_file(os.path.join(my_toolchains, '__init__.py'), "# dummy init, should not get included")

        for subdir in ['compiler', 'fft', 'linalg', 'mpi']:
            mkdir(os.path.join(my_toolchains, subdir))

        my_tc_txt = '\n'.join([
            "from easybuild.toolchains.compiler.my_compiler import MyCompiler",
            "class MyTc(MyCompiler):",
            "   pass",
        ])
        write_file(os.path.join(my_toolchains, 'my_tc.py'), my_tc_txt)

        my_compiler_txt = '\n'.join([
            "from easybuild.tools.toolchain.compiler import Compiler",
            "class MyCompiler(Compiler):",
            "   pass",
        ])
        write_file(os.path.join(my_toolchains, 'compiler', 'my_compiler.py'), my_compiler_txt)

        my_tc_bis = os.path.join(self.test_prefix, 'my_tc.py')
        write_file(my_tc_bis, '')

        # include custom toolchains
        glob_paths = [os.path.join(my_toolchains, '*.py'), os.path.join(my_toolchains, '*', '*.py'), my_tc_bis]
        included_tcs_path = include_toolchains(self.test_prefix, glob_paths)

        expected_paths = ['__init__.py', 'toolchains/__init__.py', 'toolchains/compiler/__init__.py',
                          'toolchains/my_tc.py', 'toolchains/compiler/my_compiler.py']
        for filepath in expected_paths:
            fullpath = os.path.join(included_tcs_path, 'easybuild', filepath)
            self.assertTrue(os.path.exists(fullpath), "%s exists" % fullpath)

        # path to included MNSs should be prepended to Python search path
        self.assertEqual(sys.path[0], included_tcs_path)

        # importing custom MNS should work
        import easybuild.toolchains.my_tc
        my_tc_pyc_path = easybuild.toolchains.my_tc.__file__
        my_tc_real_py_path = os.path.realpath(os.path.join(os.path.dirname(my_tc_pyc_path), 'my_tc.py'))
        self.assertTrue(os.path.samefile(up(my_tc_real_py_path, 1), my_toolchains))

    def test_is_software_specific_easyblock(self):
        """Test is_software_specific_easyblock function."""

        self.assertErrorRegex(EasyBuildError, "No such file", is_software_specific_easyblock, '/no/such/easyblock.py')

        testdir = os.path.dirname(os.path.abspath(__file__))
        test_easyblocks = os.path.join(testdir, 'sandbox', 'easybuild', 'easyblocks')

        self.assertTrue(is_software_specific_easyblock(os.path.join(test_easyblocks, 'g', 'gcc.py')))
        self.assertTrue(is_software_specific_easyblock(os.path.join(test_easyblocks, 't', 'toy.py')))

        self.assertFalse(is_software_specific_easyblock(os.path.join(test_easyblocks, 'generic', 'configuremake.py')))
        self.assertFalse(is_software_specific_easyblock(os.path.join(test_easyblocks, 'generic', 'toolchain.py')))


def suite():
    """ returns all the testcases in this module """
    return TestLoaderFiltered().loadTestsFromTestCase(IncludeTest, sys.argv[1:])


if __name__ == '__main__':
    res = TextTestRunner(verbosity=1).run(suite())
    sys.exit(len(res.failures))



##
# Copyright 2013-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for goblf compiler toolchain (includes GCC, OpenMPI, BLIS, LAPACK, ScaLAPACK and FFTW).

:author: Kenneth Hoste (Ghent University)
:author: Bart Oldeman (McGill University, Calcul Quebec, Compute Canada)
"""

from easybuild.toolchains.fft.fftw import Fftw
from easybuild.toolchains.gompi import Gompi
from easybuild.toolchains.linalg.blis import Blis
from easybuild.toolchains.linalg.lapack import Lapack
from easybuild.toolchains.linalg.scalapack import ScaLAPACK


class Goblf(Gompi, Blis, Lapack, ScaLAPACK, Fftw):
    """Compiler toolchain with GCC, OpenMPI, BLIS, LAPACK, ScaLAPACK and FFTW."""
    NAME = 'goblf'
    SUBTOOLCHAIN = Gompi.NAME



# #
# Copyright 2014-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #
"""
Unit tests for parallelbuild.py

@author: Kenneth Hoste (Ghent University)
"""
import os
import re
import stat
import sys
from test.framework.utilities import EnhancedTestCase, TestLoaderFiltered, init_config
from unittest import TextTestRunner

from easybuild.framework.easyconfig.tools import process_easyconfig
from easybuild.tools import config
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import get_module_syntax
from easybuild.tools.filetools import adjust_permissions, mkdir, read_file, remove_dir, which, write_file
from easybuild.tools.job import pbs_python
from easybuild.tools.job.pbs_python import PbsPython
from easybuild.tools.options import parse_options
from easybuild.tools.parallelbuild import build_easyconfigs_in_parallel, submit_jobs
from easybuild.tools.robot import resolve_dependencies


# test GC3Pie configuration with large resource specs
GC3PIE_LOCAL_CONFIGURATION = """[resource/ebtestlocalhost]
enabled = yes
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 1
max_memory_per_core = 1000GiB
max_walltime = 1000 hours
# this doubles as "maximum concurrent jobs"
max_cores = 1000
architecture = x86_64
auth = none
override = no
resourcedir = %(resourcedir)s
time_cmd = %(time)s
"""


MOCKED_SBATCH = """#!/bin/bash
if [[ $1 == '--version' ]]; then
    echo "slurm 17.0"
else
    echo "Submitted batch job $RANDOM"
    echo "(submission args: $@)"
fi
"""

MOCKED_SCONTROL = """#!/bin/bash
    echo "(scontrol args: $@)"
"""


def mock(*args, **kwargs):
    """Function used for mocking several functions imported in parallelbuild module."""
    return 1


class MockPbsJob(object):
    """Mocking class for PbsJob."""

    def __init__(self, *args, **kwargs):
        self.deps = []
        self.jobid = None
        self.clean_conn = None
        self.script = args[1]
        self.cores = kwargs['cores']

    def add_dependencies(self, jobs):
        self.deps.extend(jobs)

    def cleanup(self, *args, **kwargs):
        pass

    def has_holds(self, *args, **kwargs):
        pass

    def _submit(self, *args, **kwargs):
        pass


class ParallelBuildTest(EnhancedTestCase):
    """ Testcase for run module """

    def test_build_easyconfigs_in_parallel_pbs_python(self):
        """Test build_easyconfigs_in_parallel(), using (mocked) pbs_python as backend for --job."""
        # put mocked functions in place
        PbsPython__init__ = PbsPython.__init__
        PbsPython_check_version = PbsPython._check_version
        PbsPython_complete = PbsPython.complete
        PbsPython_connect_to_server = PbsPython.connect_to_server
        PbsPython_ppn = PbsPython.ppn
        pbs_python_PbsJob = pbs_python.PbsJob

        PbsPython.__init__ = lambda self: PbsPython__init__(self, pbs_server='localhost')
        PbsPython._check_version = lambda _: True
        PbsPython.complete = mock
        PbsPython.connect_to_server = mock
        PbsPython.ppn = mock
        pbs_python.PbsJob = MockPbsJob

        topdir = os.path.dirname(os.path.abspath(__file__))

        build_options = {
            'external_modules_metadata': {},
            'robot_path': os.path.join(topdir, 'easyconfigs', 'test_ecs'),
            'valid_module_classes': config.module_classes(),
            'validate': False,
            'job_cores': 3,
        }
        init_config(args=['--job-backend=PbsPython'], build_options=build_options)

        ec_file = os.path.join(topdir, 'easyconfigs', 'test_ecs', 'g', 'gzip', 'gzip-1.5-foss-2018a.eb')
        easyconfigs = process_easyconfig(ec_file)
        ordered_ecs = resolve_dependencies(easyconfigs, self.modtool)
        jobs = build_easyconfigs_in_parallel("echo '%(spec)s'", ordered_ecs, prepare_first=False)
        # only one job submitted since foss/2018a module is already available
        self.assertEqual(len(jobs), 1)
        regex = re.compile("echo '.*/gzip-1.5-foss-2018a.eb'")
        self.assertTrue(regex.search(jobs[-1].script), "Pattern '%s' found in: %s" % (regex.pattern, jobs[-1].script))

        ec_file = os.path.join(topdir, 'easyconfigs', 'test_ecs', 'g', 'gzip', 'gzip-1.4-GCC-4.6.3.eb')
        ordered_ecs = resolve_dependencies(process_easyconfig(ec_file), self.modtool, retain_all_deps=True)
        jobs = submit_jobs(ordered_ecs, '', testing=False, prepare_first=False)

        # make sure command is correct, and that --hidden is there when it needs to be
        for i, ec in enumerate(ordered_ecs):
            if ec['hidden']:
                regex = re.compile("eb %s.* --hidden" % ec['spec'])
            else:
                regex = re.compile("eb %s" % ec['spec'])
            self.assertTrue(regex.search(jobs[i].script), "Pattern '%s' found in: %s" % (regex.pattern, jobs[i].script))

        for job in jobs:
            self.assertEqual(job.cores, build_options['job_cores'])

        # no deps for GCC/4.6.3 (toolchain) and intel/2018a (test easyconfig with 'fake' deps)
        self.assertEqual(len(jobs[0].deps), 0)
        self.assertEqual(len(jobs[1].deps), 0)

        # only dependency for toy/0.0-deps is intel/2018a (dep marked as external module is filtered out)
        self.assertTrue('toy-0.0-deps.eb' in jobs[2].script)
        self.assertEqual(len(jobs[2].deps), 1)
        self.assertTrue('intel-2018a.eb' in jobs[2].deps[0].script)

        # dependencies for gzip/1.4-GCC-4.6.3: GCC/4.6.3 (toolchain) + toy/.0.0-deps
        self.assertTrue('gzip-1.4-GCC-4.6.3.eb' in jobs[3].script)
        self.assertEqual(len(jobs[3].deps), 2)
        regex = re.compile(r'toy-0.0-deps\.eb.* --hidden')
        script_txt = jobs[3].deps[0].script
        fail_msg = "Pattern '%s' should be found in: %s" % (regex.pattern, script_txt)
        self.assertTrue(regex.search(script_txt), fail_msg)
        self.assertTrue('GCC-4.6.3.eb' in jobs[3].deps[1].script)

        # also test use of --pre-create-installdir
        ec_file = os.path.join(topdir, 'easyconfigs', 'test_ecs', 't', 'toy', 'toy-0.0.eb')
        ordered_ecs = resolve_dependencies(process_easyconfig(ec_file), self.modtool)

        # installation directory doesn't exist yet before submission
        toy_installdir = os.path.join(self.test_installpath, 'software', 'toy', '0.0')
        self.assertFalse(os.path.exists(toy_installdir))

        jobs = submit_jobs(ordered_ecs, '', testing=False)
        self.assertEqual(len(jobs), 1)

        # software install dir is created (by default) as part of job submission process (fetch_step is run)
        self.assertTrue(os.path.exists(toy_installdir))
        remove_dir(toy_installdir)
        remove_dir(os.path.dirname(toy_installdir))
        self.assertFalse(os.path.exists(toy_installdir))

        # installation directory does *not* get created when --pre-create-installdir is used
        build_options['pre_create_installdir'] = False
        init_config(args=['--job-backend=PbsPython'], build_options=build_options)

        jobs = submit_jobs(ordered_ecs, '', testing=False)
        self.assertEqual(len(jobs), 1)
        self.assertFalse(os.path.exists(toy_installdir))

        # restore mocked stuff
        PbsPython.__init__ = PbsPython__init__
        PbsPython._check_version = PbsPython_check_version
        PbsPython.complete = PbsPython_complete
        PbsPython.connect_to_server = PbsPython_connect_to_server
        PbsPython.ppn = PbsPython_ppn
        pbs_python.PbsJob = pbs_python_PbsJob

    def test_build_easyconfigs_in_parallel_gc3pie(self):
        """Test build_easyconfigs_in_parallel(), using GC3Pie with local config as backend for --job."""
        try:
            import gc3libs  # noqa (ignore unused import)
        except ImportError:
            print("GC3Pie not available, skipping test")
            return

        # put GC3Pie config in place to use local host and fork/exec
        resourcedir = os.path.join(self.test_prefix, 'gc3pie')
        gc3pie_cfgfile = os.path.join(self.test_prefix, 'gc3pie_local.ini')
        gc3pie_cfgtxt = GC3PIE_LOCAL_CONFIGURATION % {
            'resourcedir': resourcedir,
            'time': which('time'),
        }
        write_file(gc3pie_cfgfile, gc3pie_cfgtxt)

        output_dir = os.path.join(self.test_prefix, 'subdir', 'gc3pie_output_dir')
        # purposely pre-create output dir, and put a file in it (to check whether GC3Pie tries to rename the output dir)
        mkdir(output_dir, parents=True)
        write_file(os.path.join(output_dir, 'foo'), 'bar')
        # remove write permissions on parent dir of specified output dir,
        # to check that GC3Pie does not try to rename the (already existing) output directory...
        adjust_permissions(os.path.dirname(output_dir), stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH,
                           add=False, recursive=False)

        topdir = os.path.dirname(os.path.abspath(__file__))

        build_options = {
            'job_backend_config': gc3pie_cfgfile,
            'job_max_walltime': 24,
            'job_output_dir': output_dir,
            'job_polling_interval': 0.2,  # quick polling
            'job_target_resource': 'ebtestlocalhost',
            'robot_path': os.path.join(topdir, 'easyconfigs', 'test_ecs'),
            'silent': True,
            'valid_module_classes': config.module_classes(),
            'validate': False,
        }
        init_config(args=['--job-backend=GC3Pie'], build_options=build_options)

        ec_file = os.path.join(topdir, 'easyconfigs', 'test_ecs', 't', 'toy', 'toy-0.0.eb')
        easyconfigs = process_easyconfig(ec_file)
        ordered_ecs = resolve_dependencies(easyconfigs, self.modtool)
        topdir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        test_easyblocks_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sandbox')
        cmd = "PYTHONPATH=%s:%s:$PYTHONPATH eb %%(spec)s -df" % (topdir, test_easyblocks_path)
        build_easyconfigs_in_parallel(cmd, ordered_ecs, prepare_first=False)

        toy_modfile = os.path.join(self.test_installpath, 'modules', 'all', 'toy', '0.0')
        if get_module_syntax() == 'Lua':
            toy_modfile += '.lua'
        self.assertTrue(os.path.exists(toy_modfile))
        self.assertTrue(os.path.exists(os.path.join(self.test_installpath, 'software', 'toy', '0.0', 'bin', 'toy')))

        # also check what happens when a job fails (an error should be raised)
        test_ecfile = os.path.join(self.test_prefix, 'test.eb')
        ectxt = read_file(ec_file)
        # use different version, for which no sources are available
        regex = re.compile('^version = .*', re.M)
        ectxt = regex.sub("version = '1.2.3'", ectxt)
        write_file(test_ecfile, ectxt)
        ecs = resolve_dependencies(process_easyconfig(test_ecfile), self.modtool)

        error = "1 jobs failed: toy-1.2.3"
        self.assertErrorRegex(EasyBuildError, error, build_easyconfigs_in_parallel, cmd, ecs, prepare_first=False)

    def test_submit_jobs(self):
        """Test submit_jobs"""
        test_easyconfigs_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'easyconfigs', 'test_ecs')
        toy_ec = os.path.join(test_easyconfigs_dir, 't', 'toy', 'toy-0.0.eb')

        args = [
            '--debug',
            '--tmpdir', '/tmp',
            '--optarch="GCC:O3 -mtune=generic;Intel:O3 -xHost"',
            '--parallel=2',
            '--try-toolchain=intel,2016a',  # should be excluded in job script
            '--robot', self.test_prefix,  # should be excluded in job script
            '--job',  # should be excluded in job script
            '--job-cores=3',
        ]
        eb_go = parse_options(args=args)
        cmd = submit_jobs([toy_ec], eb_go.generate_cmd_line(), testing=True)

        # these patterns must be found
        regexs = [
            r' --debug ',
            # values got wrapped in single quotes (to avoid interpretation by shell)
            r" --tmpdir='/tmp' ",
            r" --parallel='2' ",
            # (unparsed) optarch value got wrapped in single quotes, double quotes got stripped
            r" --optarch='GCC:O3 -mtune=generic;Intel:O3 -xHost' ",
            # templates to be completed via build_easyconfigs_in_parallel -> create_job
            r' eb %\(spec\)s ',
            r' %\(add_opts\)s ',
            r' --testoutput=%\(output_dir\)s',
            r' --disable-job ',
        ]
        for regex in regexs:
            regex = re.compile(regex)
            self.assertTrue(regex.search(cmd), "Pattern '%s' found in: %s" % (regex.pattern, cmd))

        # these patterns should NOT be found, these options get filtered out
        # (self.test_prefix was argument to --robot)
        for regex in ['--job', '--job-cores', '--try-toolchain', '--robot=[ =]', self.test_prefix + ' ']:
            regex = re.compile(regex)
            self.assertFalse(regex.search(cmd), "Pattern '%s' should *not* be found in: %s" % (regex.pattern, cmd))

    def test_build_easyconfigs_in_parallel_slurm(self):
        """Test build_easyconfigs_in_parallel(), using (mocked) Slurm as backend for --job."""

        # install mocked versions of 'sbatch' and 'scontrol' commands
        sbatch = os.path.join(self.test_prefix, 'bin', 'sbatch')
        write_file(sbatch, MOCKED_SBATCH)
        adjust_permissions(sbatch, stat.S_IXUSR, add=True)

        scontrol = os.path.join(self.test_prefix, 'bin', 'scontrol')
        write_file(scontrol, MOCKED_SCONTROL)
        adjust_permissions(scontrol, stat.S_IXUSR, add=True)

        os.environ['PATH'] = os.path.pathsep.join([os.path.join(self.test_prefix, 'bin'), os.getenv('PATH')])

        topdir = os.path.dirname(os.path.abspath(__file__))
        test_ec = os.path.join(topdir, 'easyconfigs', 'test_ecs', 'g', 'gzip', 'gzip-1.5-foss-2018a.eb')
        foss_ec = os.path.join(topdir, 'easyconfigs', 'test_ecs', 'f', 'foss', 'foss-2018a.eb')

        build_options = {
            'external_modules_metadata': {},
            'robot_path': os.path.join(topdir, 'easyconfigs', 'test_ecs'),
            'valid_module_classes': config.module_classes(),
            'validate': False,
            'job_cores': 3,
            'job_max_walltime': 5,
            'force': True,
        }
        init_config(args=['--job-backend=Slurm'], build_options=build_options)

        easyconfigs = process_easyconfig(test_ec) + process_easyconfig(foss_ec)
        ordered_ecs = resolve_dependencies(easyconfigs, self.modtool)
        self.mock_stdout(True)
        jobs = build_easyconfigs_in_parallel("echo '%(spec)s'", ordered_ecs, prepare_first=False)
        self.mock_stdout(False)

        # jobs are submitted for foss & gzip (listed easyconfigs)
        self.assertEqual(len(jobs), 2)

        # last job (gzip) has a dependency on second-to-last job (foss)
        self.assertEqual(jobs[0].job_specs['job-name'], 'foss-2018a')

        expected = {
            'dependency': 'afterok:%s' % jobs[0].jobid,
            'hold': True,
            'job-name': 'gzip-1.5-foss-2018a',
            'nodes': 1,
            'ntasks': 3,
            'output': 'gzip-1.5-foss-2018a-%j.out',
            'time': 300,  # 60*5 (unit is minutes)
            'wrap': "echo '%s'" % test_ec,
        }
        self.assertEqual(jobs[1].job_specs, expected)


def suite():
    """ returns all the testcases in this module """
    return TestLoaderFiltered().loadTestsFromTestCase(ParallelBuildTest, sys.argv[1:])


if __name__ == '__main__':
    res = TextTestRunner(verbosity=1).run(suite())
    sys.exit(len(res.failures))



##
# Copyright 2018-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
Support for using Slurm as a backend for --job

:author: Kenneth Hoste (Ghent University)
"""
import re
from distutils.version import LooseVersion

from easybuild.base import fancylogger
from easybuild.tools.build_log import EasyBuildError, print_msg
from easybuild.tools.config import JOB_DEPS_TYPE_ABORT_ON_ERROR, JOB_DEPS_TYPE_ALWAYS_RUN, build_option
from easybuild.tools.job.backend import JobBackend
from easybuild.tools.filetools import which
from easybuild.tools.run import run_cmd


_log = fancylogger.getLogger('slurm', fname=False)


class Slurm(JobBackend):
    """
    Manage SLURM server communication and create `SlurmJob` objects.
    """

    # Oldest version tested, may also work with earlier releases
    REQ_VERSION = '16.05'

    def __init__(self, *args, **kwargs):
        """Constructor."""

        # early check for required commands
        for cmd in ['sbatch', 'scontrol']:
            path = which(cmd)
            if path is None:
                raise EasyBuildError("Required command '%s' not found", cmd)

        super(Slurm, self).__init__(*args, **kwargs)

        job_deps_type = build_option('job_deps_type')
        if job_deps_type is None:
            job_deps_type = JOB_DEPS_TYPE_ABORT_ON_ERROR
            self.log.info("Using default job dependency type: %s", job_deps_type)
        else:
            self.log.info("Using specified job dependency type: %s", job_deps_type)

        if job_deps_type == JOB_DEPS_TYPE_ABORT_ON_ERROR:
            self.job_deps_type = 'afterok'
        elif job_deps_type == JOB_DEPS_TYPE_ALWAYS_RUN:
            self.job_deps_type = 'afterany'
        else:
            raise EasyBuildError("Unknown job dependency type specified: %s", job_deps_type)

    def _check_version(self):
        """Check whether version of Slurm complies with required version."""
        (out, _) = run_cmd("sbatch --version", trace=False)
        slurm_ver = out.strip().split(' ')[-1]
        self.log.info("Found Slurm version %s", slurm_ver)

        if LooseVersion(slurm_ver) < LooseVersion(self.REQ_VERSION):
            raise EasyBuildError("Found Slurm version %s, but version %s or more recent is required",
                                 slurm_ver, self.REQ_VERSION)

    def init(self):
        """
        Initialise the PySlurm job backend.
        """
        self._submitted = []

    def queue(self, job, dependencies=frozenset()):
        """
        Add a job to the queue.

        :param dependencies: jobs on which this job depends.
        """
        submit_cmd = 'sbatch'

        if dependencies:
            job.job_specs['dependency'] = self.job_deps_type + ':' + ':'.join(str(d.jobid) for d in dependencies)
            # make sure job that has invalid dependencies doesn't remain queued indefinitely
            submit_cmd += " --kill-on-invalid-dep=yes"

        # submit job with hold in place
        job.job_specs['hold'] = True

        self.log.info("Submitting job with following specs: %s", job.job_specs)
        for key in sorted(job.job_specs):
            if key in ['hold']:
                if job.job_specs[key]:
                    submit_cmd += " --%s" % key
            else:
                submit_cmd += ' --%s "%s"' % (key, job.job_specs[key])

        (out, _) = run_cmd(submit_cmd, trace=False)

        jobid_regex = re.compile("^Submitted batch job (?P<jobid>[0-9]+)")

        res = jobid_regex.search(out)
        if res:
            job.jobid = res.group('jobid')
            self.log.info("Job submitted, got job ID %s", job.jobid)
        else:
            raise EasyBuildError("Failed to determine job ID from output of submission command: %s", out)

        self._submitted.append(job)

    def complete(self):
        """
        Complete a bulk job submission.

        Release all user holds on submitted jobs, and disconnect from server.
        """
        job_ids = []
        for job in self._submitted:
            if job.job_specs['hold']:
                self.log.info("releasing user hold on job %s" % job.jobid)
                job_ids.append(job.jobid)

        if job_ids:
            run_cmd("scontrol release %s" % ' '.join(job_ids), trace=False)

        submitted_jobs = '; '.join(["%s (%s): %s" % (job.name, job.module, job.jobid) for job in self._submitted])
        print_msg("List of submitted jobs (%d): %s" % (len(self._submitted), submitted_jobs), log=self.log)

    def make_job(self, script, name, env_vars=None, hours=None, cores=None):
        """Create and return a job dict with the given parameters."""
        return SlurmJob(script, name, env_vars=env_vars, hours=hours, cores=cores)


class SlurmJob(object):
    """Job class for SLURM jobs."""

    def __init__(self, script, name, env_vars=None, hours=None, cores=None):
        """Create a new Job to be submitted to SLURM."""
        self.log = fancylogger.getLogger(self.__class__.__name__, fname=False)

        self.jobid = None
        self.script = script
        self.name = name

        self.job_specs = {
            'job-name': self.name,
            # pattern for output file for submitted job;
            # SLURM replaces %j with job ID (see https://slurm.schedmd.com/sbatch.html#lbAH)
            # %x (job name) replacement needs SLURM >= 17.02.1, thus we add name ourselves
            'output': '%s-%%j.out' % self.name,
            'wrap': self.script,
        }

        if env_vars:
            self.job_specs['export'] = ','.join(sorted(env_vars.keys()))

        max_walltime = build_option('job_max_walltime')
        if hours is None:
            hours = max_walltime
        if hours > max_walltime:
            self.log.warning("Specified %s hours, but this is impossible. (resetting to %s)" % (hours, max_walltime))
            hours = max_walltime
        self.job_specs['time'] = hours * 60

        if cores:
            self.job_specs['nodes'] = 1
            self.job_specs['ntasks'] = cores
        else:
            self.log.warning("Number of cores to request not specified, falling back to whatever Slurm does by default")



##
# Copyright 2013-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
Support for CUDA compilers as toolchain (co-)compiler.

:author: Kenneth Hoste (Ghent University)
"""

from easybuild.tools.toolchain.compiler import Compiler


TC_CONSTANT_CUDA = "CUDA"


class Cuda(Compiler):
    """CUDA compiler class."""

    COMPILER_CUDA_MODULE_NAME = ['CUDA']
    COMPILER_CUDA_FAMILY = TC_CONSTANT_CUDA

    COMPILER_CUDA_UNIQUE_OPTS = {
        # handle '-gencode arch=X,code=Y' nvcc options (also -arch, -code)
        # -arch always needs to be specified, -code is optional (defaults to -arch if missing)
        # -gencode is syntactic sugar for combining -arch/-code
        # multiple values can be specified
        # examples:
        # * target v1.3 features, generate both object code and PTX for v1.3:
        # -gencode arch=compute_13,code=compute_13 -gencode arch=compute_13,code=sm_13
        # * target v3.5 features, only generate object code for v3.5:
        # -gencode arch=compute_35,code=sm_35
        # * target v2.0 features, generate object code for v2.0 and v3.5:
        # -gencode arch=compute_20,code=sm_20 -gencode arch=compute_20,code=sm_35
        'cuda_gencode': ([], ("List of arguments for nvcc -gencode command line option, e.g., "
                              "['arch=compute_20,code=sm_20', 'arch=compute_35,code=compute_35']")),
    }

    # always C++ compiler command, even for C!
    COMPILER_CUDA_UNIQUE_OPTION_MAP = {
        '_opt_CUDA_CC': 'ccbin="%(CXX_base)s"',
        '_opt_CUDA_CXX': 'ccbin="%(CXX_base)s"',
    }

    COMPILER_CUDA_CC = 'nvcc'
    COMPILER_CUDA_CXX = 'nvcc'
    LIB_CUDA_RUNTIME = ['rt', 'cudart']

    def __init__(self, *args, **kwargs):
        """Constructor, with settings custom to CUDA."""
        super(Cuda, self).__init__(*args, **kwargs)
        # append CUDA prefix to list of compiler prefixes
        self.prefixes.append(TC_CONSTANT_CUDA)

    def _set_compiler_vars(self):
        """Set the compiler variables"""
        # append lib dir paths to LDFLAGS (only if the paths are actually there)
        root = self.get_software_root('CUDA')[0]
        self.variables.append_subdirs("LDFLAGS", root, subdirs=["lib64", "lib"])
        super(Cuda, self)._set_compiler_vars()

    def _set_compiler_flags(self):
        """Collect flags to set, and add them as variables."""

        super(Cuda, self)._set_compiler_flags()

        # always C++ compiler flags, even for C!
        # note: using $LIBS will yield the use of -lcudart in Xlinker, which is silly, but fine

        cuda_flags = [
            'Xcompiler="%s"' % str(self.variables['CXXFLAGS']),
            'Xlinker="%s %s"' % (str(self.variables['LDFLAGS']), str(self.variables['LIBS'])),
        ]
        self.variables.nextend('CUDA_CFLAGS', cuda_flags)
        self.variables.nextend('CUDA_CXXFLAGS', cuda_flags)

        # add gencode compiler flags to list of flags for compiler variables
        for gencode_val in self.options.get('cuda_gencode', []):
            gencode_option = 'gencode %s' % gencode_val
            self.variables.nappend('CUDA_CFLAGS', gencode_option)
            self.variables.nappend('CUDA_CXXFLAGS', gencode_option)



# External compatible license
# taken from https://github.com/slezica/python-frozendict on March 14th 2014 (commit ID b27053e4d1)
#
# Copyright (c) 2012 Santiago Lezica
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions
# of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE
"""
frozendict is an immutable wrapper around dictionaries that implements the complete mapping interface.
It can be used as a drop-in replacement for dictionaries where immutability is desired.
"""
import operator
from functools import reduce

from easybuild.base import fancylogger
from easybuild.tools.py2vs3 import Mapping


# minor adjustments:
# * renamed to FrozenDict
class FrozenDict(Mapping):

    def __init__(self, *args, **kwargs):
        self.__dict = dict(*args, **kwargs)
        self.__hash = None

    def __getitem__(self, key):
        return self.__dict[key]

    def copy(self, **add_or_replace):
        return FrozenDict(self, **add_or_replace)

    def __iter__(self):
        return iter(self.__dict)

    def __len__(self):
        return len(self.__dict)

    def __repr__(self):
        return '<FrozenDict %s>' % repr(self.__dict)

    def __hash__(self):
        if self.__hash is None:
            self.__hash = reduce(operator.xor, map(hash, self.iteritems()), 0)

        return self.__hash

    # minor adjustment: define missing keys() method
    def keys(self):
        return self.__dict.keys()


class FrozenDictKnownKeys(FrozenDict):
    """A frozen dictionary only allowing known keys."""

    # list of known keys
    KNOWN_KEYS = []

    def __init__(self, *args, **kwargs):
        """Constructor, only way to define the contents."""
        self.log = fancylogger.getLogger(self.__class__.__name__, fname=False)

        # support ignoring of unknown keys
        ignore_unknown_keys = kwargs.pop('ignore_unknown_keys', False)

        # handle unknown keys: either ignore them or raise an exception
        tmpdict = dict(*args, **kwargs)
        unknown_keys = [key for key in tmpdict.keys() if key not in self.KNOWN_KEYS]
        if unknown_keys:
            if ignore_unknown_keys:
                for key in unknown_keys:
                    self.log.debug("Ignoring unknown key '%s' (value '%s')" % (key, args[0][key]))
                    # filter key out of dictionary before creating instance
                    del tmpdict[key]
            else:
                msg = "Encountered unknown keys %s (known keys: %s)" % (unknown_keys, self.KNOWN_KEYS)
                self.log.raiseException(msg, exception=KeyError)

        super(FrozenDictKnownKeys, self).__init__(tmpdict)

    # pylint: disable=arguments-differ
    def __getitem__(self, key, *args, **kwargs):
        """Redefine __getitem__ to provide a better KeyError message."""
        try:
            return super(FrozenDictKnownKeys, self).__getitem__(key, *args, **kwargs)
        except KeyError as err:
            if key in self.KNOWN_KEYS:
                raise KeyError(err)
            else:
                tup = (key, self.__class__.__name__, self.KNOWN_KEYS)
                raise KeyError("Unknown key '%s' for %s instance (known keys: %s)" % tup)
